# encoding=utf-8
# Author：Wentao Zheng
# E-mail: swjtu_zwt@163.com
# developed time: 2024/2/8 12:17
import os
import socket
import struct  # 解析simulink模型打包来的数据要用
import time
import platform
import subprocess

class Client():
    def __init__(self, Send_IP='127.0.0.1', Send_Port=25001, Receive_IP='', Receive_Port=25000):
        self.send_ip = Send_IP
        self.send_port = Send_Port
        self.receive_ip = Receive_IP
        self.receive_port = Receive_Port
        self._build_client()

    def _build_client(self):
        # 发送端
        self.client_send_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # 接收端
        self.client_receive_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        
        # 尝试绑定接收端口，如果被占用则尝试其他端口
        max_attempts = 100
        current_port = self.receive_port
        for attempt in range(max_attempts):
            try:
                # 使用空字符串表示所有可用的接口
                self.client_receive_sock.bind((self.receive_ip, current_port))
                # 如果使用了不同的端口，更新实例变量
                if current_port != self.receive_port:
                    print(f"原端口{self.receive_port}被占用，已自动切换到端口{current_port}")
                    self.receive_port = current_port
                return
            except OSError as e:
                if "[WinError 10048]" in str(e) or "Address already in use" in str(e):
                    print(f"端口{current_port}已被占用，尝试使用其他端口...")
                    current_port = current_port + 1
                elif "[WinError 10013]" in str(e) or "Permission denied" in str(e):
                    # 权限问题，尝试使用更高的端口号（低于1024的端口通常需要管理员权限）
                    print(f"端口{current_port}需要管理员权限，尝试使用更高的端口号...")
                    current_port = max(1024, current_port + 1000)
                else:
                    raise e
                    
        # 如果所有尝试都失败
        raise OSError(f"无法找到可用端口，尝试了从{self.receive_port}到{current_port-1}的所有端口")

    def send_message(self, gear, acceleration, steering_angle,continue_simulation, slope=0):
        try:
            message = struct.pack('>ddddd', gear, acceleration, steering_angle,continue_simulation, slope)
            self.client_send_sock.sendto(message, (self.send_ip, self.send_port))
            if continue_simulation == 1:
                print('进入下一控制量计算过程...')
            else:
                print('控制量已发送，等到接收下一个状态量...')
        except Exception as e:
            print(f"通信出现问题！，具体原因为{e}")

    def receive_message(self):
        """接收状态数据，持续接收直到接收到结束标志"""
        received_data = []
        chunk_size = 16  # 每次接收 16 个 double 类型数据
        buffer_size = chunk_size*8  # 设置缓冲区大小
        end_flag = [-10, -10, -10, -10]  # 结束标志是 4 个 -10

        while True:
            try:
                # 接收数据
                data, addr = self.client_receive_sock.recvfrom(buffer_size)
                if not data:  # 如果没有数据可读
                    print("没有接收到数据，退出循环")
                    break

                # 解析数据长度
                num_doubles = len(data) // 8
                #print("接收到数量：",num_doubles)

                if num_doubles % 4 != 0:
                    print(f"无效数据长度: {num_doubles} doubles")
                    return None, None

                # 解包原始数据
                fmt = f'>{num_doubles}d'
                raw = struct.unpack(fmt, data)

                # 检查是否接收到结束标志
                if list(raw[-4:]) == end_flag:  # 检查最后 4 个元素是否为结束标志
                    #print("接收到结束标志，停止接收数据")
                    received_data.extend(raw[:-4])  # 排除结束标志
                    print('状态量已接收，等待计算下一个控制量...')
                    break
                else:
                    received_data.extend(raw)  # 累积接收到的数据

            except socket.error as e:
                print(f"网络错误: {e}")
                break
            except struct.error as e:
                print(f"解包错误: {e}")
                break
            except Exception as e:
                print(f"接收错误: {e}")
                break

        # 计算接收到的状态数量
        #print("receive长度",len(received_data))
        step = len(received_data) // 4

        return received_data, step


    def send_and_receive(self, gear, acceleration, steering_angle,slope):
        try:
            message = struct.pack('>dddd', gear, acceleration, steering_angle,slope)
            print("steer", steering_angle)
            self.client_send_sock.sendto(message, (self.send_ip, self.send_port))
            t1=time.time()
            print('控制量已发送，等到接收下一个状态量...')

            data, addr = self.client_receive_sock.recvfrom(1024)
            print('状态量已接收，等待计算下一个控制量...')
            t2=time.time()
            print(f"接收到数据包，耗时{t2-t1}秒")
            if data:
                unpacked_data = struct.unpack('>dddd', data)
                return unpacked_data
        except Exception as e:
            print(f"通信出现问题！，具体原因为{e}")

    def close_sockets(self):
        self.client_send_sock.close()
        self.client_receive_sock.close()
        self.kill_matlab_processes()

    @staticmethod
    def kill_matlab_processes():
        # 获取当前操作系统类型
        os_type = platform.system()

        try:
            if os_type == "Windows":
                # 对于Windows系统，使用taskkill
                subprocess.run(["taskkill", "/F", "/IM", "MATLAB.exe"], check=True)
                print("所有MATLAB进程已被成功终止。")
            elif os_type == "Linux" or os_type == "Darwin":
                # 对于Linux和MacOS系统，使用pkill
                # MacOS系统也被视为类Unix系统，通常使用和Linux相同的命令
                # subprocess.run(["pkill", "matlab"], check=True)
                os.system("ps aux|grep 'MATLAB'|grep -v 'grep'|awk '{print $2}'|xargs kill")
                print("所有MATLAB进程已被成功终止。")
            else:
                print(f"不支持的操作系统: {os_type}")
        except subprocess.CalledProcessError as e:
            print(f"终止MATLAB进程时发生错误：{e}")








import numpy as np
import matplotlib.pyplot as plt
import os
import scipy.io as sio  # 用于加载 .mat 文件
import time 

# 获取当前脚本所在的文件夹路径
current_folder = os.path.dirname(os.path.abspath(__file__))
# 获取上一级目录的路径
parent_folder = os.path.dirname(current_folder)
# 构造目标文件夹路径（例如 data 文件夹）
data_folder = os.path.join(parent_folder, 'kinetic_model')
# 构造文件的完整路径
vx_path = os.path.join(data_folder, 'vx.mat')
vy_path = os.path.join(data_folder, 'vy.mat')
grid_path = os.path.join(data_folder, 'grid.mat')

# 加载文件
vx = sio.loadmat(vx_path)
vy = sio.loadmat(vy_path)
grid = sio.loadmat(grid_path)

vx=vx['vx']
vy=vy['vy']
grid=grid['grid']



class GradientCalculator:
    def __init__(self, grid, vx, vy):
        self.grid = grid
        self.vx = vx
        self.vy = vy
        self.x_min = np.min(grid[:,:,0])
        self.y_min = np.min(grid[:,:,1])
        self.dx = 0.1
        self.dy = 0.1
        
    def get_gradient(self, x, y, theta):
        """带边界保护的梯度查询"""
        x = np.float64(x)
        y = np.float64(y)
        theta = np.float64(theta)
        
        # 计算安全索引
        i = int(np.clip(np.round((x - self.x_min)/self.dx), 
                      0, self.vx.shape[0]-1))
        j = int(np.clip(np.round((y - self.y_min)/self.dy), 
                      0, self.vx.shape[1]-1))
        dx_val = self.vx[i,j]
        dy_val = self.vy[i,j]
        return dx_val*np.cos(theta) + dy_val*np.sin(theta)



x = 1034.6#米
y = 542.7
theta = -1.566#度

calculator = GradientCalculator(grid,vx,vy)
t1 = time.time()
print("梯度值 ({:.2f},{:.2f},{:.2f}):".format(x, y, theta), 
            calculator.get_gradient(x, y, theta))

t2 = time.time()
print(t2-t1)


import os
import sys
from copy import deepcopy

class ScenarioOrganizer():
    def __init__(self):
        self.test_mode = ""
        self.scenario_list = []
        self.test_num = 0
        self.scaler = None
 
    def load(self, dir_maps:str, dir_inputs:str, dir_outputs:str) -> None:
        """读取配置文件,按照配置文件(py格式)准备场景内容
        """
        self._check_output_dir(dir_outputs)
        self.scenario_list = []
        sys.path.append(dir_inputs)
        # dir_test_conf = os.path.abspath(dir_inputs,"test_conf.py")
        from test_conf import config   
        self.config = config
        self.test_mode = config['test_settings']['mode']  # replay

        self.config['file_info'] = {
            'dir_inputs':dir_inputs,
            'dir_outputs':dir_outputs,
            'dir_scenarios':os.path.join(dir_inputs,'Scenarios'),
            'dir_maps':os.path.join(dir_inputs,'Maps'),
            'dir_maps':dir_maps,
            'location':"jiangxi_jiangtong",# default
        }

        self.config['test_settings'].setdefault('visualize',False)
         
        # 如果测试模式 == 'replay',读取文件夹下所有待测场景,存在列表中
        if self.test_mode == 'replay': # 判断测试模式是否为replay
            self.config['test_settings'].setdefault('skip_exist_scene',False)
            for item in os.listdir(self.config['file_info']['dir_scenarios']):
                scene_name_1 = item[9:-5]              
                if self.config['test_settings']['skip_exist_scene'] and os.path.exists(os.path.join(dir_outputs,scene_name_1+'_result.csv')):
                    continue
                # dir_scene_file = self.config['file_info']['dir_ScenariosResultes'] + "/" + item
                dir_scene_file = os.path.join(self.config['file_info']['dir_scenarios'], item)

                sce = self.config.copy()
                sce['data'] = {
                    'scene_name':scene_name_1,
                    'dir_scene_file':dir_scene_file
                }
                location_temp = sce['data']['scene_name'].split("_")[0]  # jiangtong/dapai
                if location_temp == "jiangtong":
                    sce['file_info']['location'] = "jiangxi_jiangtong"
                elif location_temp == "dapai":
                    sce['file_info']['location'] = "guangdong_dapai"
                elif location_temp == "shovel":
                    sce['file_info']['location'] = "anhui_shovel"
                elif 'B' in location_temp:
                    sce['file_info']['location'] = location_temp
                else:
                    raise Exception('###Exception### 地图location 错误!')
                # 将场景加入列表中
                self.scenario_list.append (deepcopy(sce))
            
            self.test_num = len(self.scenario_list)

    def next(self):
        """
        给出下一个待测场景与测试模式,如果没有场景了,则待测场景名称为None
        """
        # 首先判断测试的模式,replay模式和adaptive模式不一样
        if self.test_mode == 'replay': # 如果是回放测试
            if self.scenario_list: # 首先判断列表是否为空,如果列表不为空,则取场景;否则,输出None
                # 列表不为空,输出0号场景,且将其从列表中删除(通过pop函数实现)
                scenario_to_test = self.scenario_list.pop(0)
            else:
                # 列表为空,输出None
                scenario_to_test = None
        else:
            scenario_to_test = None
        return scenario_to_test

    def add_result(self,concrete_scenario:dict,res:float) -> None:
        # 判断测试模式,如果是replay,则忽略测试结果
        if self.test_mode == 'replay':
            return

    def _check_output_dir(self,dir_outputs:str) -> None:
        """检查输出文件夹是否存在,如果不存在,则创建

        """
        if not os.path.exists(dir_outputs):
            os.makedirs(dir_outputs)


if __name__ == "__main__":
    pass



# 内置库 
import os
import sys
import shlex
import subprocess
import numpy as np
from PIL import Image
import math
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),'..')))
import importlib
import json
# 第三方库
from typing import Dict,List,Tuple,Optional,Union

# 自定义库
from dynamic_scenes.observation import Observation
from dynamic_scenes.controller import Controller
from dynamic_scenes.recorder import Recorder
from dynamic_scenes.visualizer import Visualizer
from dynamic_scenes.lookup import CollisionLookup
from dynamic_scenes.kinetics_model import KineticsModelStarter

class Env():
    """仿真环境读取及迭代过程,simulation"""
    def __init__(self):
        self.controller = Controller()
        self.recorder = Recorder()
        self.visualizer = Visualizer()
        self.scene_type = "intersection"
    def get_slope(self, observation, height):
        z1 = height[20][20]
        dx = observation['vehicle_info']['ego']['shape']['length']*np.cos(observation['vehicle_info']['ego']['yaw_rad'])
        dy = observation['vehicle_info']['ego']['shape']['length']*np.sin(observation['vehicle_info']['ego']['yaw_rad'])
        x_idx = math.floor(dx+20)
        y_idx = math.floor(dy+20)
        x_idx = max(0, min(x_idx, 40))
        y_idx = max(0, min(y_idx, 40))
        z2 = height[x_idx][y_idx]

        distance = math.sqrt(dx ** 2 + dy ** 2)
        return -math.atan2((z2-z1), distance)

    def init_height_loader(self,scenario:dict, dir_maps, dir_inputs, scene_type='intersection') -> None:

        # 战区高程读取逻辑
        if scene_type == "B":
            self.location_temp = scenario['data']['scene_name'].split("_")[0]  # Bxxx
            # height_file_path = os.path.abspath(os.path.join(dir_inputs, 'Maps', 'bitmap', self.location_temp+'_global_map.json'))
            height_file_path = os.path.abspath(os.path.join(dir_maps, 'bitmap', self.location_temp+'_global_map.json'))
            try:
                with open(height_file_path, 'r', encoding='utf-8') as height_file:
                    mission_height_json = json.load(height_file)
                data = mission_height_json.get('data')
                # battle_height是二维数组，查询方式是self.battle_height[y_pix][x_pix],y_pix和x_pix分别是高方向和宽方向上的像素索引，一般是物理坐标/resolution
                self.battle_height_info = [[-1 if item == "-1" else item for item in row] for row in data]
                self.battle_height = mission_height_json['metadata']['dimensions']['height']
                self.battle_width = mission_height_json['metadata']['dimensions']['width']
                self.battle_reso = mission_height_json['metadata']['dimensions']['resolution']

            except:
                print(f"### fail to load height data from {height_file_path}!")
        else:
            self.location_temp = scenario['data']['scene_name'].split("_")[0]  # jiangtong/dapai/shovel
            dir_current_file = os.path.dirname(__file__)  
            dir_parent_1 = os.path.dirname(dir_current_file) 
            data_path = os.path.abspath(os.path.join(dir_parent_1, 'data'))
            height_file_path = os.path.abspath(os.path.join(data_path, self.location_temp + '_block.npy'))
            try:
                self.points = np.load(height_file_path)
                self.points_dict = {(row[0], row[1]): row[2] for row in self.points}
            except:
                print(f"### fail to load height data from {height_file_path}!")

            # map_path = os.path.join(dir_inputs,'Maps')
            # semanticmap_path = os.path.join(map_path,'semantic_map')
            semanticmap_path = os.path.join(dir_maps,'semantic_map')
            for item in os.listdir(semanticmap_path):
                if self.location_temp in item:
                    file_path = os.path.join(semanticmap_path, item)
                    if os.path.isfile(file_path):
                        try:
                            # 打开并加载 JSON 文件
                            with open(file_path, 'r', encoding='utf-8') as file:
                                data = json.load(file)
                            print(f"成功加载文件：{file_path}")
                            # 在这里处理 data
                        except json.JSONDecodeError:
                            print(f"错误：文件 {file_path} 不是有效的 JSON 格式！")
                        except Exception as e:
                            print(f"加载文件 {file_path} 时出错：{e}")
                    self.para1  = -data["bitmap_mask_PNG"]["UTM_info"]["point_southwest"][0]
                    self.para2 = data["bitmap_mask_PNG"]["canvas_edge_meter"][1]
                    self.para3 = data["bitmap_mask_PNG"]["UTM_info"]["point_southwest"][1]
        



    def make(self,scenario:dict, read_only=False, dir_maps='', dir_inputs='', save_img_path='',kinetics_mode='simple', scene_type = "intersection") -> Tuple:
        """第一次进入读取环境信息.

        Args:
            scenario (dict): 动态场景输入信息.
            collision_lookup (CollisionLookup): 用于 ego车与(栅格化mask)边界 进行碰撞检测的预置数据.
            read_only (bool, optional): _description_. Defaults to False.

        Returns:
            Observation: 当前时刻环境观察结果;
            traj:全局的背景车辆轨迹数据;
        """
        self.scene_type = scene_type
        # 载入高程信息
        self.init_height_loader(scenario, dir_maps, dir_inputs, self.scene_type)

        observation,observation_all,traj = self.controller.init(scenario,kinetics_mode, self.scene_type)
        if kinetics_mode == "complex":
            Starter = KineticsModelStarter(observation)
            self.client = Starter.get_client
            print("using complex kinetics model")
        else:
            raise ValueError("暂不提供这种动力学模式，请选择complex！")

        self.recorder.init(observation_all,scenario['file_info']['dir_outputs'],read_only)
        if self.scene_type == "B":
            self.visualizer.battle_init(observation_all,
                            scenario['test_settings']['visualize'],
                            scenario['test_settings']['save_fig_whitout_show'],
                            img_save_path=save_img_path) # 此处通过查看配置参数,True,设置运行过程中可视化打开;
        else:
            self.visualizer.init(observation_all,
                            scenario['test_settings']['visualize'],
                            scenario['test_settings']['save_fig_whitout_show'],
                            img_save_path=save_img_path) # 此处通过查看配置参数,True,设置运行过程中可视化打开;
        
        height = self.get_height(observation.format())
        
        return observation.format(),self.client, height
    
    def get_height(self, observation):
        result = []
        if self.scene_type=="B":
            for i in range(-20,21):
                col = []
                for j in range(-20, 21):
                    x = observation['vehicle_info']['ego']['x'] + i - observation['test_setting']['x_min']
                    y = observation['vehicle_info']['ego']['y'] + j - observation['test_setting']['y_min']
                    x_pix = int(x/self.battle_reso)
                    y_pix = int(y/self.battle_reso)
                    x_pix = min((len(self.battle_height_info[0])-1), max(x_pix, 0))
                    y_pix = min((len(self.battle_height_info)-1), max(y_pix, 0)) 
                    z_data = self.battle_height_info[y_pix][x_pix]
                    col.append(z_data)
                result.append(col)
        else:
            for i in range(-20, 21):  # 遍历 -20 到 20
                col = []
                for j in range(-20, 21):  # 遍历 -20 到 20
                    unround_x = (observation['vehicle_info']['ego']['x'] + i) + self.para1
                    unround_y = self.para2 - (observation['vehicle_info']['ego']['y'] + j) + self.para3
                    position_transformed = (round(int(unround_x)+0.5, 1), round(int(unround_y)+0.5, 1))  # 构造查询键

                    # 查询字典，如果不存在则返回 -1
                    z_data = self.points_dict.get(position_transformed, -1)
                    col.append(z_data)  # 将高度值添加到当前列
                result.append(col)  # 将当前列添加到结果列表
        return result
        

    def step(self,action:Tuple[float,float,int],traj_future:Dict,observation_last:Observation,traj=200, path_planned=None, vis_cir_open=False) -> Observation:
        """迭代过程"""
        last_height = self.get_height(observation_last)
        last_slope = self.get_slope(observation_last,last_height)
        observation = self.controller.step(self,action,self.client,traj_future,observation_last,traj, path_planned, last_slope, last_height=None, vis_cir_open=vis_cir_open)  # 使用车辆运动学模型单步更新场景;
        height = self.get_height(observation.format())
        #self.recorder.record(observation)
        # self.visualizer.update(observation)  
        #self.visualizer.update(observation,traj_future,observation_last,traj)# 更新场景后,使用更新的ego车辆位置进行可视化;【CZF】添加预测轨迹+真实轨迹的比较
        
        return observation.format(), height



if __name__ == "__main__":
    import time
    dir_current_file = os.path.dirname(__file__)  
    dir_parent_1 = os.path.dirname(dir_current_file) 
    height_data_path = os.path.abspath(os.path.join(dir_parent_1, 'data'))
    print(dir_parent_1)
    points = np.load(os.path.join(height_data_path,'shovel_block.npy'))
    input_image_path = os.path.join(height_data_path, 'anhui_shovel_bitmap_mask.png')
    output_image_path = os.path.join(height_data_path, 'anhui_shovel_change.png')
    img = Image.open(input_image_path)
    img = img.convert('RGB')  # 转换为RGB模式以便着色
    img_array = np.array(img)  # 转换为numpy数组
    # 获取图像的宽度和高度
    width, height = img.size

    # # 打印宽度和高度
    print(f"图像的宽度: {width} 像素")
    print(f"图像的高度: {height} 像素")


    # # 转换物理坐标到像素坐标
    # x_pixels = (points[:, 0] / 0.1).astype(int)  # x轴坐标转换
    # y_pixels = (points[:, 1] / 0.1).astype(int)  # y轴坐标转换


    # 创建有效坐标掩码（防止越界）
    # valid_mask = (
    #     (x_pixels >= 0) & 
    #     (x_pixels < width) & 
    #     (y_pixels >= 0) & 
    #     (y_pixels < height)
    # )

    # # 应用掩码过滤无效坐标
    # x_valid = x_pixels[valid_mask]
    # y_valid = y_pixels[valid_mask]

    # # 在图像数组上绘制红色标记点（RGB格式）
    # img_array[y_valid, x_valid] = [255, 0, 0]  # 红色
    x_car = 729.5
    y_car = 1275.5
    data_dict = {(row[0], row[1]): row[2] for row in points}

    print(x_car, y_car)
    for i in range(-20,20):
        for j in range(-20,20):
            y = int(y_car+j)+0.5
            x = int(x_car+i)+0.5
            test_x_y = (x, y)
            if test_x_y in data_dict:
                img_array[int(y*10), int(x*10)] = [0, 255, 0]
            else:
                img_array[int(y*10), int(x*10)] = [255, 0, 0]


            
    # 保存处理后的图像
    result_img = Image.fromarray(img_array)
    
    result_img.show()





# 内置库
import os
import time

# 第三方库
import json
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
# from matplotlib.axes import Axes
# from matplotlib.figure import Figure
# from matplotlib.patches import Rectangle,Arrow
# from scipy.ndimage import zoom
from scipy.signal import convolve2d
from matplotlib.colors import LinearSegmentedColormap
from typing import Dict, List, Tuple, Optional, Union, Any

Image.MAX_IMAGE_PIXELS = None
Axis = Any


# plt.style.use('seaborn-whitegrid')# Seaborn 库提供的一种基于白色网格背景的样式.


class BitMap:
    """BitMap类,用于读取、显示栅格地图等功能.
    矿区地图的信息包括以下栅格图层(rasterized):
    1)可行驶区域图.png,bitmap_mask.png
        - 江西-江铜:"jiangtong";
        - 广东-大排:"jiangtong";        
    使用与Google Maps/Earth 相同的WGS 84 Web Mercator (EPSG:3857);"CRS_name":"GCS_WGS_1984_UTM_Zone_50N".
    2) bitmap_rgb.png
        由航空制图转化为不带坐标信息的rgb栅格图,可直观显示整个区域全貌.
    """

    def __init__(self,
                 dataroot: str = '/home/czf/project_czf/20231010_onsite_mine/datasets',
                 location: str = "jiangxi_jiangtong",
                 bitmap_type: str = 'bitmap_mask',
                 is_transform_gray=False,
                 scene_type=''):
        """地图中的位图(栅格图)包括:
         - bitmap_mask:语义先验(可行驶区域)黑白图掩码(mask)图层, bitmap_mask.png文件 .
         - bitmap_rgb :将航空制图转化为PNG彩色图, bitmap_rgb.png文件.

        Args:
            dataroot (str,optional):datasets根目录. Defaults to '/home/czf/project_czf/20231010_onsite_mine/datasets'.
            location (str,optional):具体矿区位置的名称,当前有以下可选 :`jiangxi_jiangtong`,`guangdong_dapai`. Defaults to "jiangxi_jiangtong".
            bitmap_type (str,optional):位图类型,当前可选:`bitmap_mask`,`bitmap_rgb`..Defaults to 'bitmap_mask'.
            is_transform_gray (bool,optional):黑白图 OR 灰色图. Defaults to False.
            self.bitmap_info是整个地图的信息，self.bitmap_local_info是场景中局部地图信息，self.image_ndarray_local为局部像素矩阵
        """

        self.dataroot = dataroot
        self.location = location
        self.bitmap_type = bitmap_type

        if scene_type == "B":
            self.load_battlefield_info()
        else:
            self.load_bitmap_info()
            self.image_ndarray = self.load_bitmap(is_transform_gray=is_transform_gray)

    def load_battlefield_info(self):
        dir_semantic_map = os.path.join(self.dataroot, "bitmap", self.location + '_global_map.json')
        with open(dir_semantic_map, 'r') as f:
            temp_semantic_map_info = json.load(f)
        # 加载信息
        # 读取的地图文件中记录的长宽应该是像素意义上的长宽，指该方向上有几个像素
        xmin = temp_semantic_map_info['metadata']['origin'][0]
        ymin = temp_semantic_map_info['metadata']['origin'][1]
        xmax = xmin + temp_semantic_map_info['metadata']['dimensions']["width"]*temp_semantic_map_info['metadata']['dimensions']["resolution"]
        ymax = ymin + temp_semantic_map_info['metadata']['dimensions']["height"]*temp_semantic_map_info['metadata']['dimensions']["resolution"]
        xrange = xmax - xmin
        yrange = ymax - ymin
        pxrange = int(temp_semantic_map_info['metadata']['dimensions']["width"])
        pyrange = int(temp_semantic_map_info['metadata']['dimensions']["height"])
        self.bitmap_info = {}
        temp_bitmap_info = {
            "png_version":"shovel_mask.png,20231214",
            "UTM_info":{
                "point_southwest":[xmin,ymin],
                "point_northeast":[xmax,ymax],
                "local_x_range":[xmin,xmax],
                "local_y_range":[ymin,ymax],
                "unit":"m"
            },
            "canvas_edge_meter":[xrange,yrange],
            "canvas_edge_pixel":[pxrange,pyrange],
            "scale_PixelPerMeter":1/temp_semantic_map_info['metadata']['dimensions']["resolution"],
            "scale_MeterPerPixel":0.1
        }
        # self.bitmap_info['bitmap_rgb_PNG'] = temp_semantic_map_info['bitmap_rgb_PNG']
        self.bitmap_info['bitmap_mask_PNG'] = temp_bitmap_info

        data = np.array(temp_semantic_map_info.get('data'))
        width = temp_semantic_map_info['metadata']['dimensions']["width"]
        height = temp_semantic_map_info['metadata']['dimensions']["height"]
        image_ndarray = np.ones((height,width),np.uint8)*255
        for i in range(height):
            for j in range(width):
                if(data[i][j]=="-1"):
                    image_ndarray[i, j] = 0
        # image_ndarray的i，j用的与物理坐标系相同，即(x,y)对应的是image_ndarray[y/resolution, x/resolution]
        self.image_ndarray = image_ndarray

    def load_bitmap_info(self):
        """load bitmap information.
        """
        semantic_map_hashes = {
            "jiangxi_jiangtong": 'jiangxi_jiangtong_semantic_map',
            "guangdong_dapai": 'guangdong_dapai_semantic_map',
            "anhui_shovel": 'anhui_shovel_semantic_map'
        }
        semantic_map_hash = semantic_map_hashes[self.location]
        dir_semantic_map = os.path.join(self.dataroot, "semantic_map", semantic_map_hash + '.json')
        with open(dir_semantic_map, 'r') as f:
            temp_semantic_map_info = json.load(f)
        self.bitmap_info = {}
        self.bitmap_info['bitmap_rgb_PNG'] = temp_semantic_map_info['bitmap_rgb_PNG']
        self.bitmap_info['bitmap_mask_PNG'] = temp_semantic_map_info['bitmap_mask_PNG']

    def load_bitmap(self, is_transform_gray=False) -> np.ndarray:
        """ 加载指定的 bitmap.
         
        Args:
            is_transform_gray (bool, optional): 黑白图 OR 灰色图. Defaults to False.

        Returns:
            np.ndarray: 栅格图数组(全局).
        """

        if self.bitmap_type == 'bitmap_mask':
            bitmap_hashes = {
                "jiangxi_jiangtong": 'jiangxi_jiangtong_bitmap_mask',
                "guangdong_dapai": 'guangdong_dapai_bitmap_mask',
                "anhui_shovel": 'anhui_shovel_bitmap_mask'
            }
            bitmap_hash = bitmap_hashes[self.location]
            dir_bitmap = os.path.join(self.dataroot, 'bitmap', bitmap_hash + '.png')
            if os.path.exists(dir_bitmap):
                image = Image.open(dir_bitmap)
                image_ndarray = np.array(image.convert('1'))  # 转为 np.ndarray
                image_ndarray = np.flipud(image_ndarray)  # 翻转Y轴
                # contains_true = np.any(image_ndarray)
                # colors = [(0.5, 0.5, 0.5), (1, 1, 1)]  # R -> G -> B
                # n_bins = [3]  # Discretizes the interpolation into bins
                # cmap_name = 'custom_div_cmap'
                # cm = LinearSegmentedColormap.from_list(cmap_name, colors, N=2)
                # fig = plt.figure(figsize=(10.0, 10.0))
                # ax = fig.add_subplot()
                # ax.imshow(image_ndarray, extent=(self.bitmap_info['bitmap_mask_PNG']['UTM_info']['local_x_range'][0],
                #                                  self.bitmap_info['bitmap_mask_PNG']['UTM_info']['local_x_range'][1],
                #                                  self.bitmap_info['bitmap_mask_PNG']['UTM_info']['local_y_range'][0],
                #                                  self.bitmap_info['bitmap_mask_PNG']['UTM_info']['local_y_range'][1]), cmap=cm, origin='lower')
                # plt.show()
                # image_ndarray = image_ndarray.max() - image_ndarray # 实现颜色反转的效果.
            else:
                raise Exception('###Exception### %s的地图路径不存在 %s! Please check.' % (self.bitmap_type, dir_bitmap))

        elif self.bitmap_type == 'bitmap_rgb':
            bitmap_hashes = {
                "jiangxi_jiangtong": "jiangxi_jiangtong_bitmap_rgb",
                "guangdong_dapai": "guangdong_dapai_bitmap_rgb"}
            bitmap_hash = bitmap_hashes[self.location]
            dir_bitmap = os.path.join(self.dataroot, 'bitmap', bitmap_hash + '.png')
            if os.path.exists(dir_bitmap):
                if is_transform_gray == False:
                    image = Image.open(dir_bitmap)
                    image_ndarray = np.array(image.convert('RGB'))
                else:
                    image = Image.open(dir_bitmap)
                    image_ndarray = np.array(image.convert('L'))
            else:
                raise Exception('###Exception### %s的地图路径不存在 %s! Please check.' % (self.bitmap_type, dir_bitmap))

        else:
            raise Exception('###Exception### 非法的 bitmap type:%s' % self.bitmap_type)  # 自定义异常

        self.dir_bitmap = dir_bitmap

        return image_ndarray

    def load_bitmap_using_utm_local_range(self,
                                          utm_local_range: Tuple[float, float, float, float] = (0.0, 0.0, .1, .1),
                                          x_margin=10, y_margin=10, is_battle = False):
        """
        根据局部坐标范围获取对应的局部区域的像素矩阵
        utm_local_range:The rectangular patch using utm local coordinates (x_min,y_min,x_max,y_max).
         

        Args:
            utm_local_range (Tuple[float,float,float,float], optional): The rectangular patch 
                using utm local coordinates (x_min,y_min,x_max,y_max).Defaults to (0.0,0.0,.1,.1).
            x_margin (int, optional): x外边缘尺寸. Defaults to 10m.
            y_margin (int, optional): y外边缘尺寸. Defaults to 10m.

        Returns:
            np.ndarray: 栅格图数组(局部).
        """

        self.bitmap_local_info = dict()
        self.bitmap_local_info["utm_local_range"] = utm_local_range
        self.bitmap_local_info["x_margin"] = x_margin
        self.bitmap_local_info["y_margin"] = y_margin
        # calculate pixel  
        px_min, py_min = self._utm_to_pixel(utm_local_range[0] - x_margin,
                                            utm_local_range[1] - y_margin)  # 注意像素坐标与UTM坐标在y轴上的方向映射是正确的
        px_max, py_max = self._utm_to_pixel(utm_local_range[2] + x_margin, utm_local_range[3] + y_margin)
        self.bitmap_local_info["pixel_local_range"] = (px_min, py_min, px_max, py_max)

        # 修正像素尺寸为方形
        px_diff = px_max - px_min
        py_diff = py_max - py_min
        max_diff = max(px_diff, py_diff)
        px_max = px_min + max_diff
        py_max = py_min + max_diff

        # 使用像素坐标从self.image_ndarray中切片出局部矩阵
        self.image_ndarray_local = self.image_ndarray[py_min:py_max, px_min:px_max]

    def _utm_to_pixel(self, x, y):
        """将UTM坐标(x,y)转换为像素坐标(pixel_x,pixel_y).
        """

        x_range = self.bitmap_info['bitmap_mask_PNG']['UTM_info']['local_x_range']
        y_range = self.bitmap_info['bitmap_mask_PNG']['UTM_info']['local_y_range']
        scale = self.bitmap_info['bitmap_mask_PNG']['scale_PixelPerMeter']

        # 转换坐标
        pixel_x = int((x - x_range[0]) * scale)
        pixel_y = int((y - y_range[0]) * scale)  # 注意y轴已经反过来了

        return pixel_x, pixel_y  # 返回的是像素点的索引（右下或者左上）

    def render_mask_map_using_image_ndarray_local(self, ax: Axis = None, window_size=4, gray_flag=True):
        """ 渲染可行驶区域二值化栅格图(局部);
        注:黑白图 和 灰色图 两种.        
         
        Args:
            ax (Axis,optional):坐标轴ax. Defaults to None.
            window_size (int,optional):降采样窗口大小,使用均值滤波器. Defaults to 4.
            gray_flag (bool,optional):黑白图 OR 灰色图. Defaults to True.
        """
        if self.bitmap_type != 'bitmap_mask':
            raise ValueError('Error:render_mask_map()函数只有使用semantic_prior图初始化bitmap才可以调用!')

        if ax is None:
            ax = plt.subplot()

        # 滤波+降采样
        image_ndarray_local = self.downsample_with_filter(self.image_ndarray_local, window_size)

        # 根据降采样的比率调整UTM的范围
        x_min = self.bitmap_local_info['utm_local_range'][0] - self.bitmap_local_info['x_margin']
        y_min = self.bitmap_local_info['utm_local_range'][1] - self.bitmap_local_info['y_margin']
        x_max = self.bitmap_local_info['utm_local_range'][2] + self.bitmap_local_info['x_margin']
        y_max = self.bitmap_local_info['utm_local_range'][3] + self.bitmap_local_info['y_margin']

        if gray_flag == True:  # 黑白图
            ax.imshow(image_ndarray_local, extent=(x_min, x_max, y_min, y_max), cmap='gray', origin='lower')
        else:  # 创建从灰到白的自定义 colormap
            colors = [(0.5, 0.5, 0.5), (1, 1, 1)]  # R -> G -> B
            n_bins = [3]  # Discretizes the interpolation into bins
            cmap_name = 'custom_div_cmap'
            cm = LinearSegmentedColormap.from_list(cmap_name, colors, N=2)
            ax.imshow(image_ndarray_local, extent=(x_min, x_max, y_min, y_max), cmap=cm, origin='lower')

    @staticmethod
    def downsample_with_filter(image: np.ndarray = None, window_size: int = 5) -> np.ndarray:
        """均值滤波器对图像进行降采样.

        Args:
            image (np.ndarray, optional): 原图像数组. Defaults to None.
            window_size (int, optional): 滤波窗口尺寸. Defaults to 5.

        Returns:
            np.ndarray: 滤波后图像数组.
        """

        # 定义一个均值滤波器
        kernel = np.ones((window_size, window_size), dtype=np.float32) / (window_size * window_size)
        # 使用滤波器对图像进行滤波
        filtered_image = convolve2d(image.astype(float), kernel, mode='same', boundary='wrap')
        # 降采样
        downsampled = filtered_image[::window_size, ::window_size]
        # 二值化
        downsampled_binary = (downsampled > 0.5).astype(np.bool)

        return downsampled_binary

    def render_mask_map(self, ax: Axis = None, gray_flag: bool = True):
        """ 渲染可行驶区域二值化栅格图(全局图像),semantic_prior.
        注1:整个图都会被渲染,只是显示范围由其它调用函数决定;注2:黑白图 和 灰色图 两种.  

        Args:
            ax (Axis,optional):可选参数,用于指定渲染图像的轴(坐标轴). Defaults to None.
            gray_flag (bool,optional):黑白图 OR 灰色图. Defaults to True.
        """

        assert self.bitmap_type == 'bitmap_mask', 'Error:render_mask_map()函数只有使用bitmap_mask图初始化bitmap才可以调用!'

        if ax is None:
            ax = plt.subplot()

        # 当前map的实际地理尺寸大小,(宽,高)米 . meters (width,height).
        x_min, x_max = self.UTM_local_x_range
        y_min, y_max = self.UTM_local_y_range
        ax.imshow(self.image_ndarray, extent=[x_min, x_max, y_min, y_max], cmap='gray', origin='lower')

        if gray_flag == True:  # 黑白图
            ax.imshow(self.image_ndarray, extent=[x_min, x_max, y_min, y_max], cmap='gray', origin='lower')
        else:  # 创建从灰到白的自定义 colormap
            colors = [(0.5, 0.5, 0.5), (1, 1, 1)]  # R -> G -> B
            n_bins = [3]  # Discretizes the interpolation into bins
            cmap_name = 'custom_div_cmap'
            cm = LinearSegmentedColormap.from_list(cmap_name, colors, N=2)
            ax.imshow(self.image_ndarray, extent=[x_min, x_max, y_min, y_max], cmap=cm, origin='lower')

    def render_rgb_map(self, ax: Axis = None):
        """渲染栅格图中的 bitmap_rgb.png
 
        Args:
            ax (Axis,optional):可选参数,用于指定渲染图像的轴(坐标轴). Defaults to None.
        """

        assert self.bitmap_type == 'bitmap_rgb', 'Error:render_rgb_map()函数只有使用GeoTIFF_2_PNG图初始化bitmap才可以调用!'

        if ax is None:
            ax = plt.subplot()
        x, y = self.canvas_edge_meter
        if len(self.image_ndarray.shape) == 2:
            ax.imshow(self.image_ndarray, extent=[0, x, 0, y], cmap='gray')
        else:
            ax.imshow(self.image_ndarray, extent=[0, x, 0, y])



# 导入内置库
import sys
import os
import time

# 导入第三方库
import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("TkAgg")
from typing import Dict, List, Tuple, Optional, Union, Any
from matplotlib.patches import Rectangle, Arrow
from matplotlib.axes import Axes
from PIL import Image
import io


# # 添加必要的路径到sys.path
def add_path_to_sys(target: str):
    abs_path = os.path.abspath(target)
    if abs_path not in sys.path:
        sys.path.append(abs_path)


dir_current_file = os.path.dirname(__file__)  # onsite_mine_plus/devkit/map_expansion/  
dir_parent_1 = os.path.dirname(dir_current_file)  # onsite_mine_plus/devkit/
dir_parent_2 = os.path.dirname(dir_parent_1)  # onsite_mine/
dir_parent_3 = os.path.dirname(dir_parent_2)  # onsite_mine/..
add_path_to_sys(dir_current_file)
add_path_to_sys(dir_parent_1)
add_path_to_sys(os.path.join(dir_parent_1, "common"))
add_path_to_sys(os.path.join(dir_parent_1, "dynamic_scenes"))
add_path_to_sys(os.path.join(dir_parent_1, "map_expansion"))

# 自定义库
import common.utils as utils
from map_expansion.map_api import TgScenesMap
from map_expansion.map_api import TgScenesMapExplorer
from map_expansion.bit_map import BitMap


class GlobalMapVisualizer:

    def __init__(self, dataroot: str = "onsite_mine/datasets", location: str = "jiangxi_jiangtong"):
        self.flag_visilize = True  # 是\否进行可视化绘图
        self.flag_hdmaps_visilized = False  # hdmaps已经被绘图(仅绘制一次即可)标志位
        self.flag_visilize_ref_path = True  # 是否可视化参考路径
        self.flag_visilize_prediction = True  # 是否可视化轨迹预测结果
        self.flag_save_fig_whitout_show = False  # 绘图保存,但是不显示

        self.dataroot = dataroot
        self.location = location

        self.hdmaps = {}
        self.hdmaps = self._parse_hdmaps()
        self.color_map = dict(
            road="#b2df8a",
            road_block="#2A2A2A",
            intersection="#fb9a99",
            loading_area="#6a3d9a",
            unloading_area="#7e772e",
            connector_path="#1f78b4",
            base_path="#1f78b4",
            dubins_pose="#064b7a",
        )

    def _parse_hdmaps(self):
        """解析高清地图文件.
        功能:
        1)获取可行驶区域的mask图信息.
        2)加载路网结构.
        """

        # 获取mask图信息并确定最大方形区域
        self.hdmaps["image_mask"] = BitMap(self.dataroot, self.location, "bitmap_mask")
        self.x_min = self.hdmaps["image_mask"].bitmap_info["bitmap_mask_PNG"]["UTM_info"]["point_southwest"][0]
        self.x_max = self.hdmaps["image_mask"].bitmap_info["bitmap_mask_PNG"]["UTM_info"]["point_northeast"][0]
        self.y_min = self.hdmaps["image_mask"].bitmap_info["bitmap_mask_PNG"]["UTM_info"]["point_southwest"][1]
        self.y_max = self.hdmaps["image_mask"].bitmap_info["bitmap_mask_PNG"]["UTM_info"]["point_northeast"][1]

        # 加载路网信息
        self.hdmaps["tgsc_map"] = TgScenesMap(self.dataroot, self.location)
        self.hdmaps = {"image_mask": self.hdmaps["image_mask"], "tgsc_map": self.hdmaps["tgsc_map"]}

        return self.hdmaps

    def init(self, fig_size: float = 100):
        """绘图"""
        self.fig_size = fig_size
        # 初始化画布
        if self.flag_visilize:
            if self.flag_save_fig_whitout_show == True:
                plt.ioff()
            else:
                plt.ion()
            pixel_size = self.hdmaps["image_mask"].bitmap_info["bitmap_mask_PNG"]["canvas_edge_pixel"]  # width,height
            # width_normalize, height_normalize = width / max(width, height), height / max(width, height)
            dpi = 100  # 每英寸的像素点数
            # 系统没有足够的内存来分配给图像缓冲区。这个问题通常发生在尝试创建非常大的图像时，比如当图像尺寸或分辨率特别高时。
            fig_size = [pixel_size[0] / dpi, pixel_size[1] / dpi]  # 以英寸为单位的画布大小

            # figsize_calculate = [width_normalize* fig_size ,   height_normalize* fig_size ]  # 获取比例关系
            self.fig = plt.figure(figsize=fig_size)
            self.axbg = self.fig.add_subplot()  # 多个x轴,共用y轴
            # self.axveh = self.axbg.twiny()
            # self.axTraj = self.axbg.twiny()
        else:
            plt.ioff()
            return

        self.flag_hdmaps_visilized = False  # 本次绘制地图
        if self.hdmaps:
            self._plot_hdmaps()

    def add_token(self):
        pass

    def _plot_hdmaps(self, padding: float = 20):
        # 1) 绘制地图个图层; 2)标记 各元素的token;
        if not self.hdmaps:
            return
        # plot mask图,plot 道路片段、交叉口多边形划分;
        my_patch = (self.x_min - padding, self.y_min - padding, self.x_max + padding, self.y_max + padding)  #  (x_min,y_min,x_max,y_max).
        # layer_names = ['intersection','road']
        layer_names = ["road", "intersection", "loading_area", "unloading_area", "road_block"]
        self._plot_hdmaps_render_map_patch(
            tgsc_map_explorer=TgScenesMapExplorer(self.hdmaps["tgsc_map"]),
            box_coords=my_patch,
            layer_names=layer_names,
            alpha=0.7,
            bitmap=self.hdmaps["image_mask"],
            ax=self.axbg,
        )

    def _plot_hdmaps_render_map_patch(
        self,
        tgsc_map_explorer: TgScenesMapExplorer = None,
        box_coords: Tuple[float, float, float, float] = (0, 0, 1, 1),
        layer_names: List[str] = None,
        alpha: float = 0.5,
        bitmap: Optional[BitMap] = None,
        ax: Axes = None,
    ) -> None:
        """渲染一个矩形图框,指定矩形框的xy坐标范围.By default renders all layers."""

        if bitmap is not None:
            if bitmap.bitmap_type == "bitmap_mask":
                bitmap.render_mask_map(ax, gray_flag=False)

        for layer_name in layer_names:  # 渲染各图层
            tgsc_map_explorer._render_layer(ax, layer_name, alpha)

        #  渲染intersection的reference path,lane
        if self.flag_visilize_ref_path == True:
            tgsc_map_explorer.render_connector_path_centerlines(ax, alpha, resampling_rate=0.05)

        #  渲染road的reference path,lane
        if self.flag_visilize_ref_path == True:
            tgsc_map_explorer.render_base_path_centerlines(ax, alpha, resampling_rate=0.2)


def save_fig(filename):
    # 保存图像
    buf = io.BytesIO()
    plt.savefig(buf, format="png")
    buf.seek(0)
    img = Image.open(buf)
    img.save(filename)
    buf.close()


def view_global_map_demo():
    """示例,查看and保存全局地图."""
    dir_datasets = os.path.abspath(os.path.join(dir_parent_3, "datasets"))  # 指定数据集根目录
    location = 'guangdong_dapai'#  指定矿区,索引对应地图
    # location = "jiangxi_jiangtong"  #  指定矿区,索引对应地图
    # 1 解析地图
    global_map_visualizer = GlobalMapVisualizer(dataroot=dir_datasets, location=location)

    # 2 初始化 图幅,应该根据全局地图的尺寸自动初始化一个长宽;
    global_map_visualizer.init()

    # 3绘制各图层的token
    global_map_visualizer.add_token()

    # 自动保存 save 到当前目录下
    dir_datasets_parent= os.path.dirname(dir_datasets)
    # --------刷新当前帧画布--------
    # plt.show()
    plt.subplots_adjust()
    save_fig("global_map2.png")


if __name__ == "__main__":
    # Parse and view the global map
    view_global_map_demo()



from enum import Enum, unique
import math
import time


@unique
class VehicleType(Enum):
    MineTruck_XG90G = 1  # 车辆类型:徐工宽体车90
    MineTruck_NTE200 = 2  # 车辆类型:北方股份电动论矿卡NTE200


class MineTruckXG90G:
    def __init__(self):
        # 车辆最大外形轮廓
        self.length = 9.0
        self.width = 4.0
        self.locationPoint2Head = 6.5
        self.locationPoint2Rear = 2.5
        # 道路边界碰撞检测框轮廓
        self.collision_l = 9.0
        self.collision_w = 4.0
        self.collision_lrear = 6.5
        self.collision_lhead = 2.5


class MineTruckNTE200:
    def __init__(self):
        # 车辆最大外形轮廓
        self.length = 13.0
        self.width = 6.7
        self.locationPoint2Head = 9.2
        self.locationPoint2Rear = 3.8
        # 道路边界碰撞检测框轮廓
        self.collision_l = 8.8
        self.collision_w = 5.5
        self.collision_lrear = 1.6
        self.collision_lhead = 7.2


def sign(x):
    if x >= 0:
        return 1
    else:
        return -1


class Point2D:
    """2D平面的位姿点."""

    def __init__(self, x, y, yaw_rad: float = 0.0):
        self.x = x
        self.y = y
        self.yaw = yaw_rad


class CollisionLookup:
    """道路边界碰撞检测.
       通过建立一个查找表(lookup table)，能够快速检测给定位置和方向的车辆是否会与环境中的障碍物边界发生碰撞.
       注：与其它车辆进行碰撞检测不按照该方法.
    ## 1.碰撞查找表的生成过程
        首先，根据车辆尺寸和预设的参数，初始化车辆在不同位置和方向上的可能状态。
        对于每个可能的位置和方向，计算车辆四个角的位置，并根据这些角的位置来填充一个称为C空间的二维数组。C空间反映了车辆所占的空间。
        然后，利用直线扫描算法（Bresenham算法的一个变种），标记出所有可能与车辆发生碰撞的栅格。
        最后，将这些数据保存到查找表中，供碰撞检测时查询使用。
    ## 2.碰撞检测过程
        给定一个车辆位姿(x，y,yaw) 和地图，函数首先将这些参数转换为查找表中的索引。
        然后，遍历对应的查找表项，检查车辆所占的每个栅格是否与地图上的障碍物重叠。
        如果存在重叠（即发现碰撞），则立即返回碰撞标志。
    ## 3.性能优化
        这个系统的设计允许快速碰撞检测，避免了复杂计算和直接的几何碰撞检测算法。
        通过预计算和存储车辆在不同位置和方向上的占用格，显著减少了实时计算的需要，适用于需要快速响应的自动驾驶和机器人导航系统。
    """

    def __init__(self, type: VehicleType = VehicleType.MineTruck_XG90G):
        """构造函数初始化一系列参数，如车辆尺寸、碰撞检测精度、方向的离散化等，并计算出用于碰撞检测的查找表."""
        # !初始化车辆参数
        if type == VehicleType.MineTruck_NTE200:
            self.vehicle = MineTruckXG90G()
        elif type == VehicleType.MineTruck_XG90G:
            self.vehicle = MineTruckXG90G()
        else:
            self.vehicle = MineTruckXG90G()

        # !初始化碰撞检测精度参数
        self._headings = 72  # 空间分辨率,方向的离散化程度,即在全圆中考虑的不同方向的数量.
        self._c_size = 0.2  # 碰撞检测的空间分辨率,单位是米.
        self._position_resolution = 1  # _c_size*_c_size 格子进一步细化
        self._positions = self._position_resolution**2
        self.expansion_dis = 0.0  # 碰撞尺寸,单位米;
        # 基于车辆尺寸和self._c_size参数计算的边界框大小，用于定义碰撞检测时考虑的空间区域
        self._bbSize = math.ceil((math.sqrt(self.vehicle.collision_w**2 + (self.vehicle.collision_l * 2) ** 2) + self.expansion_dis) / self._c_size)
        self._delt_heading_rad = 3.1415926 * 2 / self._headings
        self._delt_heading_rad = 3.1415926 * 2 / self._headings

        # 生成每个栅格的离散点位
        self._points = [Point2D(0.0, 0.0) for i in range(self._positions)]
        for i in range(self._position_resolution):
            for j in range(self._position_resolution):
                pt = Point2D(1.0 / self._position_resolution * j, 1.0 / self._position_resolution * i)
                self._points[i * self._position_resolution + j] = pt
        self._c = Point2D(0.0, 0.0)  # 车辆中心
        self._p = [Point2D(0.0, 0.0) for i in range(4)]  # 车辆的四个角
        self._np = [Point2D(0.0, 0.0) for i in range(4)]  # 旋转后车辆的四个角
        self._lookup = [list() for i in range(self._positions * self._headings)]  # 表格

        for q in range(self._positions):
            theta = 0.0
            # 计算车辆中心
            self._c.x = self._bbSize / 2.0 + self._points[q].x
            self._c.y = self._bbSize / 2.0 + self._points[q].y
            # 计算车辆四个角位置
            self._p[0].x = self._c.x - self.vehicle.collision_lrear / self._c_size
            self._p[0].y = self._c.y - self.vehicle.collision_w / 2 / self._c_size
            self._p[1].x = self._c.x - self.vehicle.collision_lrear / 2 / self._c_size
            self._p[1].y = self._c.y + self.vehicle.collision_w / 2 / self._c_size
            self._p[2].x = self._c.x + self.vehicle.collision_lhead / self._c_size
            self._p[2].y = self._c.y + self.vehicle.collision_w / 2 / self._c_size
            self._p[3].x = self._c.x + self.vehicle.collision_lhead / self._c_size
            self._p[3].y = self._c.y - self.vehicle.collision_w / 2 / self._c_size

            for o in range(self._headings):
                self._cSpace = [False for i in range(self._bbSize**2)]  # 初始化C空间

                # 旋转变换
                for i in range(4):
                    temp_x = self._p[i].x - self._c.x
                    temp_y = self._p[i].y - self._c.y
                    self._np[i].x = temp_x * math.cos(theta) - temp_y * math.sin(theta) + self._c.x
                    self._np[i].y = temp_x * math.sin(theta) + temp_y * math.cos(theta) + self._c.y
                theta += self._delt_heading_rad

                # 单元格顺时针遍历
                for i in range(4):
                    start = Point2D(0, 0)
                    end = Point2D(0, 0)
                    # 起始点和终点设置
                    if i < 3:
                        start = self._np[i]
                        end = self._np[i + 1]
                    else:
                        start = self._np[i]
                        end = self._np[0]

                    # x,y向下取整得到下标
                    x_id = math.floor(start.x)
                    y_id = math.floor(start.y)
                    self._cSpace[y_id * self._bbSize + x_id] = True  # 占用

                    # 搜索起点到终点的方向向量
                    t = Point2D(end.x - start.x, end.y - start.y)
                    # 下一步移动方向(按格点移动)
                    step_x = sign(t.x)
                    step_y = sign(t.y)
                    # 下一步移动距离占总移动距离的比例(x,y方向)
                    t_delta_x = 1000
                    t_delta_y = 1000
                    if t.x != 0:
                        t_delta_x = 1.0 / abs(t.x)  # 一个格点占x移动距离的比例
                    if t.y != 0:
                        t_delta_y = 1.0 / abs(t.y)  # 一个格点占y移动距离的比例
                    t_max_x, t_max_y = 0.0, 0.0
                    if step_x > 0:
                        t_max_x = t_delta_x * (1 - (start.x - math.trunc(start.x)))
                    else:
                        t_max_x = t_delta_x * (start.x - math.trunc(start.x))
                    if step_y > 0:
                        t_max_y = t_delta_y * (1 - (start.y - math.trunc(start.y)))
                    else:
                        t_max_y = t_delta_y * (start.y - math.trunc(start.y))

                    # 边界占用栅格搜索
                    while math.floor(end.x) != x_id or math.floor(end.y) != y_id:
                        # 当x所占比例小,且移动后距离终点更近时,移动x
                        if t_max_x < t_max_y and abs(x_id + step_x - math.floor(end.x)) < abs(x_id - math.floor(end.x)):
                            t_max_x += t_delta_x
                            x_id += step_x
                            self._cSpace[y_id * self._bbSize + x_id] = True
                        elif t_max_y < t_max_x and abs(y_id + step_y - math.floor(end.y)) < abs(y_id - math.floor(end.y)):
                            t_max_y += t_delta_y
                            y_id += step_y
                            self._cSpace[y_id * self._bbSize + x_id] = True
                        elif 2 >= abs(x_id - math.floor(end.x)) + abs(y_id - math.floor(end.y)):
                            if abs(x_id - math.floor(end.x)) > abs(y_id - math.floor(end.y)):
                                x_id += step_x
                                self._cSpace[y_id * self._bbSize + x_id] = True
                            else:
                                y_id += step_y
                                self._cSpace[y_id * self._bbSize + x_id] = True
                        else:
                            print("\n--->tie occured,please check for error in script\n")
                            break

                # 填满图形
                hcross1, hcross2 = 0, 0
                for i in range(self._bbSize):
                    inside = False
                    # 寻找覆盖区域的起点和终点
                    for k in range(self._bbSize):
                        if self._cSpace[i * self._bbSize + k] == True and inside == False:
                            hcross1 = k
                            inside = True
                        elif self._cSpace[i * self._bbSize + k] == True and inside == True:
                            hcross2 = k
                    # 如果没有找到覆盖区域
                    if inside == False:
                        continue
                    # 填充覆盖区域
                    for j in range(self._bbSize):
                        if j >= hcross1 and j <= hcross2:
                            self._cSpace[i * self._bbSize + j] = True

                # 设置表格
                for i in range(self._bbSize):
                    for j in range(self._bbSize):
                        if self._cSpace[i * self._bbSize + j]:
                            self._lookup[q * self._headings + o].append(Point2D(j - math.floor(self._c.x), i - math.floor(self._c.y)))

    def collision_detection(self, x: float, y: float, yaw: float, image):
        """接收位置(x，y)、偏航角和一个二维地图，判断在该位置和方向上的车辆是否会碰撞.

        Args:
            x (float): x,y——与地图原点的相对坐标
            y (float): x,y——与地图原点的相对坐标
            yaw (float): 车辆偏航角
            image (np.array): 二值化地图(表示是否被占用)

        Returns:
            bool: 是否与边界发生碰撞.
        """
        start_time = time.perf_counter()

        x = x / 0.1  # self._c_size
        y = y / 0.1  # self._c_size
        X = math.floor(x)
        Y = math.floor(y)
        iX = math.floor((x - X) * self._position_resolution)
        iY = math.floor((y - Y) * self._position_resolution)
        while yaw > math.pi * 2:
            yaw -= math.pi * 2
        while yaw < 0:
            yaw += math.pi * 2
        iH = math.floor(yaw / self._delt_heading_rad)
        iH = min((self._headings - 1), max(0, iH))
        idx = iY * self._position_resolution * self._headings + iX * self._headings + iH
        idx = min((len(self._lookup) - 1), max(0, idx))

        for i in range(len(self._lookup[idx])):
            cX = X + self._lookup[idx][i].x * 2
            cY = Y + self._lookup[idx][i].y * 2
            if cX < 0 or cX >= image.shape[1] or cY < 0 or cY >= image.shape[0]:
                return True
            if image[cY][cX] == False:
                return True
        end_time = time.perf_counter()
        # print("##log##车辆与道路边界碰撞检测时间:"+str(end_time-start_time))
        return False



# 内置库 
import math
import os
import sys
import copy
from itertools import combinations
from typing import Tuple,Optional
import time

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),'..')) )
from typing import Dict,List,Tuple,Optional,Union
# 第三方库
import json
import numpy as np
from shapely.geometry import Polygon

# 自定义库
import common.utils as utils
from map_expansion.map_api import TgScenesMap
from map_expansion.bit_map import BitMap
from dynamic_scenes.lookup import CollisionLookup,VehicleType
from dynamic_scenes.observation import Observation
from dynamic_scenes.socket_module import Client




class ReplayInfo():
    """用于存储回放测试中用以控制背景车辆的所有数据
        背景车轨迹信息 vehicle_traj
        vehicle_traj = {
            "vehicle_id_0":{
                "shape":{
                    "vehicle_type":"PickupTruck",
                    "length":5.416,
                    "width":1.947,
                    "height":1.886,
                    "locationPoint2Head":2.708,
                    "locationPoint2Rear":2.708
                },
                "t_0":{
                    "x":0,
                    "y":0,
                    "v_mps":0,
                    "acc_mpss":0,
                    "yaw_rad":0,
                    "yawrate_radps":0,
                },
                "t_1":{...},
                ...
            },
            "vehicle_id_1":{...},
            ...
        }
        主车轨迹信息,只包含当前帧信息
        ego_info = {
            "shape":{
                        "vehicle_type":"MineTruck_NTE200",
                        "length":13.4,
                        "width":6.7,
                        "height":6.9,
                        "min_turn_radius":14.2,
                        "locationPoint2Head":2.708,
                        "locationPoint2Rear":2.708
                    },
            "x":0,
            "y":0,
            "v_mps":0,
            "acc_mpss":0,
            "yaw_rad":0,
            "yawrate_radps":0
        }
        地图相关信息,具体介绍地图解析工作的教程 markdown 文档待编写
        self.hdmaps = {
            'image_mask':bitmap的ndarray,
            'tgsc_map':语义地图
            }
        测试环境相关信息 test_setting
        test_setting = {
            "t":,
            "dt":,
            "max_t",
            "goal":{
                "x":[-1,-1,-1,-1],
                "y":[-1,-1,-1,-1]
            },
            "end":,
            "scenario_type":,
            "enter_loading_flag":,
            "enter_loading_time":,
            "scenario_name":,
            "map_type":,
            "start_ego_info" 
        }

    """


    def __init__(self):
        self.vehicle_traj = {}  # 背景车轨迹
        self.ego_info = {
            "shape":{
                        "vehicle_type":"MineTruck_NTE200",
                        "length":13.4,
                        "width":6.7,
                        "height":6.9,
                        "min_turn_radius":14.2,
                        "locationPoint2Head":2.708,
                        "locationPoint2Rear":2.708
                    },
            "x":0,
            "y":0,
            "v_mps":0,
            "acc_mpss":0,
            "yaw_rad":0,
            "yawrate_radps":0,
        }
        self.hdmaps = {}
        self.test_setting = {
            "t":0,
            "dt":0.1,
            "max_t":10,
            "goal":{
                "x":[1,2,3,4 ],
                "y":[1,2,3,4 ],
                "head": None
            },# goal box:4 points [x1,x2,x3,x4],[y1,y2,y3,y4]
            "end":-1,
            "scenario_name":None,
            "scenario_type":None,
            "enter_loading_flag":False,
            "enter_loading_time":0.00,
            "x_min":None,
            "x_max":None,
            "y_min":None,
            "y_max":None,
            "start_ego_info":None  # 用不到，不会读取
        } #同 Observation.test_setting


    def _add_vehicle_shape(self, id: int, traj_info: dict = None):
        """ 
        该函数实现向vehicle_trajectiry中添加背景车轨迹信息的功能——增加车辆形状
        """
        if id not in self.vehicle_traj.keys():
            self.vehicle_traj[id] = {}
            self.vehicle_traj[id]['shape'] = {}
        self.vehicle_traj[id]['shape']= traj_info['VehicleShapeInfo']
     
            
    def _add_vehicle_traj(self,id: int, traj_info: dict = None):
        """ 
        该函数实现向vehicle_trajectiry中添加背景车轨迹信息的功能-增加车辆状态、轨迹
        """
        for index,_ in enumerate(traj_info['states']['x']):
            t = traj_info['StartTimeInScene']+ index * self.test_setting['dt']
            str_t = str(round(float(t),2))  # 注意这里的key为时间，是str
            if str_t not in self.vehicle_traj[id].keys(): 
                self.vehicle_traj[id][str_t]={}
            for key,value in zip(['x','y','yaw_rad','v_mps','yawrate_radps','acc_mpss'],
                                  [traj_info['states']['x'][index][0],traj_info['states']['y'][index][0],
                                   traj_info['states']['yaw_rad'][index][0],traj_info['states']['v_mps'][index][0],
                                   traj_info['states']['yawrate_radps'][index][0],traj_info['states']['acc_mpss'][index][0]]):
                if value is not None:
                    self.vehicle_traj[id][str_t][key] = (np.around(value,5))  # 轨迹保留5位小数

    @staticmethod
    def normalize_angle(angle):
        normalized_angle = angle % (2 * math.pi)
        return normalized_angle
    
    
    def add_ego_info(self,ego_info):
        """
        该函数实现向test_setting中添加自车初始状态信息
        """
        self.ego_info['x'] = ego_info['states']['x']
        self.ego_info['y'] = ego_info['states']['y']
        self.ego_info['yaw_rad'] = self.normalize_angle(ego_info['states']['yaw_rad'])
        self.ego_info['v_mps'] = ego_info['states']['v_mps']
        self.ego_info['yawrate_radps'] = ego_info['states']['yawrate_radps']
        self.ego_info['acc_mpss'] = ego_info['states']['acc_mpss']
        self.ego_info['shape'] = ego_info['VehicleShapeInfo']
        
        
    def _add_settings(self,scenario_name=None,scenario_type=None):
        """
        该函数实现向test_setting中添加場景名稱和場景類型
        """
        for key,value in zip(['scenario_name','scenario_type'],
                              [scenario_name,scenario_type]):
            if value is not None:
                self.test_setting[key] = value

    def _init_battlefield_vehicle_ego_info(self,one_scenario: dict = None):
        self.ego_info['shape'] = one_scenario['ego_info']['VehicleShapeInfo']
        self.ego_info['x'] = one_scenario['ego_info']['initial_state']['x']
        self.ego_info['y'] = one_scenario['ego_info']['initial_state']['y']
        self.ego_info['yaw_rad'] = one_scenario['ego_info']['initial_state']['yaw_rad']
        self.ego_info['v_mps'] = one_scenario['ego_info']['initial_state']['v_mps']
        self.ego_info['acc_mpss'] = one_scenario['ego_info']['initial_state']['yawrate_radps']
        self.ego_info['yawrate_radps'] = one_scenario['ego_info']['initial_state']['acc_mpss']

    def _init_vehicle_ego_info(self,one_scenario: dict = None):
        """
        该函数实现向ego_info中增加主车信息的功能
        注意:ego_info中只含有主车当前帧的信息
        """
        self.ego_info['shape'] = one_scenario['ego_info']['VehicleShapeInfo']
        self.ego_info['x'] = one_scenario['ego_info']['start_states']['x']
        self.ego_info['y'] = one_scenario['ego_info']['start_states']['y']
        self.ego_info['yaw_rad'] = one_scenario['ego_info']['start_states']['yaw_rad']
        self.ego_info['v_mps'] = one_scenario['ego_info']['start_states']['v_mps']
        self.ego_info['acc_mpss'] = one_scenario['ego_info']['start_states']['acc_mpss']
        self.ego_info['yawrate_radps'] = one_scenario['ego_info']['start_states']['yawrate_radps']


    def _get_dt_maxt(self,one_scenario: dict = None):
        """
        该函数实现得到最大仿真时长阈值以及采样率的功能,最大最小xy范围
        """
        self.test_setting['max_t']= one_scenario['max_t']
        self.test_setting['dt']= one_scenario['dt']
        self.test_setting['x_min']= one_scenario['x_min']
        self.test_setting['x_max']= one_scenario['x_max']
        self.test_setting['y_min']= one_scenario['y_min']
        self.test_setting['y_max']= one_scenario['y_max']

    def _get_goal(self,one_scenario: dict = None):
        x_A1 = one_scenario['goal']['x'][0]
        x_A2 = one_scenario['goal']['x'][1]
        x_A3 = one_scenario['goal']['x'][2]
        x_A4 = one_scenario['goal']['x'][3]

        y_A1 = one_scenario['goal']['y'][0]
        y_A2 = one_scenario['goal']['y'][1]
        y_A3 = one_scenario['goal']['y'][2]
        y_A4 = one_scenario['goal']['y'][3]
        self.test_setting['goal'] = {
                                        "x": [x_A1,x_A2,x_A3,x_A4],
                                        "y": [y_A1,y_A2,y_A3,y_A4],
                                        "heading": one_scenario['goal']['heading']
                                    }
        
        x = np.mean(self.test_setting['goal']["x"])
        y = np.mean(self.test_setting['goal']["y"])
        self.test_setting['point_goal'] = [x,y,np.mean(self.test_setting['goal']["heading"][0])]



class ReplayParser():
    """
    解析场景文件
    """
    def __init__(self):
        self.replay_info = ReplayInfo()


    def parse(self,scenario:dict, scene_type) -> ReplayInfo:
        """解析动态场景.动态场景包括: 
        1) 自车信息,self.replay_info.ego_info;
        2) 高精度地图信息,self.replay_info.hdmaps;
        3) 测试场景信息,self.replay_info.test_setting;
        4) 其它背景车辆全部轨迹信息,self.replay_info.vehicle_traj;
        
        Args:
            scenario (dict): 动态场景输入信息.

        Returns:
            ReplayInfo: 动态场景信息解析结果.
        """
        # 场景名称与测试类型
        self.scene_type = scene_type
        if scene_type == "B":
            self.replay_info._add_settings(scenario_name=scenario['data']['scene_name'],
                            scenario_type=scenario['data']['scene_name'].split("_")[1])
            dir_scene_file = scenario['data']['dir_scene_file']
            self._parse_battlefield_scenario(dir_scene_file)
            self._parse_battle_field_hdmaps(scenario) # 解析 地图文件
        else:
            self.replay_info._add_settings(scenario_name=scenario['data']['scene_name'],
                                        scenario_type=scenario['data']['scene_name'].split("_")[1])
            dir_scene_file = scenario['data']['dir_scene_file']
            self._parse_scenario(dir_scene_file) # 解析多车场景
            self._parse_hdmaps(scenario) # 解析 地图文件
        
        return self.replay_info
    
    def _parse_battlefield_scenario(self, dir_scene_file):
        with open(dir_scene_file,'r') as f:
            one_scenario = json.load(f)

        # 1) 获取ego车辆的目标区域,goal box
        # self.replay_info.test_setting['goal'] = one_scenario['goal']
        # self.replay_info.test_setting['goal'].setdefault('head', None)
        self.replay_info._get_goal(one_scenario)
        # 2) 步长,最大时间,最大最小xy范围
        self.replay_info._get_dt_maxt(one_scenario)
        
        # 3) 读取ego车初始信息
        self.replay_info._init_battlefield_vehicle_ego_info(one_scenario)

        # 读取背景车信息，包括：车辆形状信息以及轨迹信息
        for idex,value_traj_segment in enumerate(one_scenario['TrajSegmentInfo']):
            # if value_traj_segment['TrajSetToken'] != "ego":
            num_vehicle = idex
            # 4) 读取车辆长度与宽度等形状信息,录入replay_info.id从1开始
            self.replay_info._add_vehicle_shape(
                id=num_vehicle,
                traj_info=value_traj_segment)
            # 5) 以下读取背景车相关信息,车辆编号从1号开始,轨迹信息记录在vehicle_traj中
            self.replay_info._add_vehicle_traj(
                id=num_vehicle,
                traj_info=value_traj_segment)

        return self.replay_info

    def _parse_scenario(self, dir_scene_file:str):
        """解析多车场景.
        """
        with open(dir_scene_file,'r') as f:
            one_scenario = json.load(f)

        # 1) 获取ego车辆的目标区域,goal box
        self.replay_info.test_setting['goal'] = one_scenario['goal']
        self.replay_info.test_setting['goal'].setdefault('head', None)
        
        # 2) 步长,最大时间,最大最小xy范围
        self.replay_info._get_dt_maxt(one_scenario)
        
        # 3) 读取ego车初始信息
        self.replay_info._init_vehicle_ego_info(one_scenario)

        # 读取背景车信息，包括：车辆形状信息以及轨迹信息
        for idex,value_traj_segment in enumerate(one_scenario['TrajSegmentInfo']):
            # if value_traj_segment['TrajSetToken'] != "ego":
            num_vehicle = idex +1
            # 4) 读取车辆长度与宽度等形状信息,录入replay_info.id从1开始
            self.replay_info._add_vehicle_shape(
                id=num_vehicle,
                traj_info=value_traj_segment)
            # 5) 以下读取背景车相关信息,车辆编号从1号开始,轨迹信息记录在vehicle_traj中
            self.replay_info._add_vehicle_traj(
                id=num_vehicle,
                traj_info=value_traj_segment)

        return self.replay_info
    
    def _parse_battle_field_hdmaps(self,scenario:dict) -> None:
        dataroot = scenario['file_info']['dir_maps']
        location = scenario['file_info']['location']
        
        # 获取mask图信息并确定最大方形区域
        self._load_battlefied_mask_and_calculate_square_region(dataroot,location)
    
    def _load_battlefied_mask_and_calculate_square_region(self,dataroot:str,location:str) -> None:
        self.replay_info.hdmaps['image_mask'] = BitMap(dataroot,location,'bitmap_mask', scene_type=self.scene_type)  # 得到整个二进制图的ndarray
        
        # 整合所有的坐标点
        x_coords = (
            [self.replay_info.ego_info['x']] + 
            self.replay_info.test_setting['goal']['x'] +
            [self.replay_info.test_setting['x_min'],self.replay_info.test_setting['x_max']]
        )  # 包括主车初始横坐标、目标区域横坐标、以及整张地图横坐标
        y_coords = (
            [self.replay_info.ego_info['y']] + 
            self.replay_info.test_setting['goal']['y'] +
            [self.replay_info.test_setting['y_min'],self.replay_info.test_setting['y_max']]
        )
        
        # 根据坐标确定最大的方形框
        x_min,x_max = self.replay_info.test_setting['x_min'], self.replay_info.test_setting['x_max']
        y_min,y_max = self.replay_info.test_setting['y_min'], self.replay_info.test_setting['y_max']
        utm_local_range = (
            x_min,y_min,
            x_max,y_max
        )  # 确定坐标范围
        self.replay_info.hdmaps['image_mask'].load_bitmap_using_utm_local_range(utm_local_range,0,0, is_battle=True)
   
    def _parse_hdmaps(self,scenario:dict) -> None:
        """解析高清地图文件并更新到replay_info.
        Args: 这里的scenario为下一个待测场景列表中下一个待测场景对应的dict
        功能:
        1)获取可行驶区域的mask图信息.
        2)获取rgb图的目录(用于可视化) .
        3)加载路网结构.
        """
        # 初始化相关路径
        dataroot = scenario['file_info']['dir_maps']
        location = scenario['file_info']['location']
        
        # 1) 获取mask图信息并确定最大方形区域
        self._load_mask_and_calculate_square_region(dataroot,location)
        
        # 2) 获取rgb图信息 (  如果需要 )
        # self._load_rgb_image(dataroot,location)
        
        # 3) 加载路网信息
        self._load_road_network(dataroot,location)
        
        
    def _load_mask_and_calculate_square_region(self,dataroot:str,location:str) -> None:
        """加载mask并确定最大的方形区域.
            Args:
                dataroot: maps的根目录位置
                location: jiangxi_jiangtong/guangdong_dapai
        """
        self.replay_info.hdmaps['image_mask'] = BitMap(dataroot,location,'bitmap_mask')  # 得到整个二进制图的ndarray
        
        # 整合所有的坐标点
        x_coords = (
            [self.replay_info.ego_info['x']] + 
            self.replay_info.test_setting['goal']['x'] +
            [self.replay_info.test_setting['x_min'],self.replay_info.test_setting['x_max']]
        )  # 包括主车初始横坐标、目标区域横坐标、以及整张地图横坐标
        y_coords = (
            [self.replay_info.ego_info['y']] + 
            self.replay_info.test_setting['goal']['y'] +
            [self.replay_info.test_setting['y_min'],self.replay_info.test_setting['y_max']]
        )
        
        # 根据坐标确定最大的方形框
        x_min,x_max = min(x_coords),max(x_coords)
        y_min,y_max = min(y_coords),max(y_coords)
        max_range = max(x_max - x_min,y_max - y_min)  # 确定正方形边长
        x_center,y_center = (x_min + x_max) / 2,(y_min + y_max) / 2  # 确定正方形中心
        utm_local_range = (
            x_center - max_range / 2,y_center - max_range / 2,
            x_center + max_range / 2,y_center + max_range / 2
        )  # 确定坐标范围
        self.replay_info.hdmaps['image_mask'].load_bitmap_using_utm_local_range(utm_local_range,20,20)  # 希望加载的位图范围能比坐标范围大一点
        # 局部位图二进制范围结果储存在self.replay_info.hdmaps['image_mask'].image_ndarray_local中
        
    
    def _load_rgb_image(self,dataroot:str,location:str) -> None:
        """加载地图的RGB图像.
        """
        self.replay_info.hdmaps['image_rgb'] = BitMap(dataroot,location,'bitmap_rgb',is_transform_gray=True)
    
    
    def _load_road_network(self,dataroot:str,location:str) -> None:
        """加载地图的路网信息.
        """
        self.replay_info.hdmaps['tgsc_map'] = TgScenesMap(dataroot,location)  #语义地图信息
        
          
        
        

class ReplayController():
    def __init__(self, kinetics_mode):
        self.control_info = ReplayInfo()
        self.collision_lookup = None
        self.kinetics_mode = kinetics_mode
        self.step_count = None  # 添加一个类属性来存储step_count


    def init(self,control_info:ReplayInfo) -> Observation:
        self.control_info = control_info
        # self.collision_lookup = collision_lookup
        return self._get_initial_observation()

    def step(self,env,action,old_observation:Observation,traj_future:Dict,observation_last:Observation,traj=200, ha_path=None, client:Optional[Client] = None, slope=0, last_height=None, vis_cir_open=False) -> Observation:
        """
        进行一步更新
        new_observation:限制范围内观测值
        new_observation_all:全部观测值用于可视化
        """
        if self.kinetics_mode == 'complex':
            action = self._action_cheaker_kinetics(action)
            new_observation,new_observation_all = self._update_kinetics(env,action,old_observation, traj_future,observation_last,traj,client, ha_path, slope, last_height, vis_cir_open)
            #print("完成一次更新与记录")
        else:
            action = self._action_cheaker(action)
            new_observation = self._update_ego_and_t(action, old_observation)
            new_observation = self._update_other_vehicles(new_observation)
            new_observation = self._update_end_status(new_observation)
            new_observation_all=new_observation

        return new_observation,new_observation_all


    def _action_cheaker(self,action):
        a = np.clip(action[0], -15, 15)
        rad = np.clip(action[1], -1, 1)
        gear = action[2]
        if gear not in [1, 2, 3]:
            raise ValueError("不支持{0}档位，请选择合适档位".format(gear))
        return (a,rad,gear)

    def _action_cheaker_kinetics(self,action):
        a = np.clip(action[0],-4,2)
        rad = np.clip(action[1],-math.pi/2,math.pi/2)
        gear = action[2]
        if gear not in [1, 2, 3]:
            raise ValueError("不支持{0}档位，请选择合适档位".format(gear))
        return (a,rad,gear)


    def _get_initial_observation(self) -> Observation:
        observation = Observation()
        # vehicle_info
        ego_vehicle = self.control_info.ego_info
        observation.vehicle_info["ego"] = ego_vehicle  # 初始化主车信息

        # 更新自车与边界碰撞查询表
        if ego_vehicle["shape"]["length"] >= 8.5 and ego_vehicle["shape"]["width"] <= 9.5:
            self.collision_lookup = CollisionLookup(type=VehicleType.MineTruck_XG90G)
        elif ego_vehicle["shape"]["length"] >= 12.5 and ego_vehicle["shape"]["width"] <= 13.6:
            self.collision_lookup = CollisionLookup(type=VehicleType.MineTruck_NTE200)
        else:
            self.collision_lookup = CollisionLookup(type=VehicleType.MineTruck_XG90G)
        # hdmaps info
        observation.hdmaps = self.control_info.hdmaps
        # test_setting
        observation.test_setting = self.control_info.test_setting
        observation,observation_all = self._update_other_vehicles(observation)  # 初始化背景车信息

        observation = self._update_end_status(observation)
        observation_all = self._update_end_status(observation_all)
    
        return observation,observation_all


    def _update_ego_and_t(self,action:tuple,old_observation:Observation) -> Observation:
        # 拷贝一份旧观察值
        new_observation = copy.copy(old_observation)
        # 首先修改时间,新时间=t+dt
        new_observation.test_setting['t'] = float(
            old_observation.test_setting['t'] +
            old_observation.test_setting['dt']
        )
        # 修改本车的位置,方式是前向欧拉更新,1.根据旧速度更新位置;2.然后更新速度.
        # 速度和位置的更新基于自行车模型.
        # 首先分别取出加速度和方向盘转角
        a,rot,gear = action
        # 取出步长
        dt = old_observation.test_setting['dt']
        # 取出本车的各类信息
        x,y,v,yaw,= [float(old_observation.vehicle_info['ego'][key]) for key in ['x','y','v_mps','yaw_rad']]
        width,length  = old_observation.vehicle_info['ego']['shape']['width'],old_observation.vehicle_info['ego']['shape']['length']

        # 首先根据旧速度更新本车位置
        new_observation.vehicle_info['ego']['x'] = x + \
                                                   v * np.cos(yaw) * dt  # 更新X坐标

        new_observation.vehicle_info['ego']['y'] = y + \
                                                   v * np.sin(yaw) * dt  # 更新y坐标

        new_observation.vehicle_info['ego']['yaw_rad'] = yaw + \
                                                     v / length * 1.7 * np.tan(rot) * dt  # 更新偏航角

        new_observation.vehicle_info['ego']['v_mps'] = v + a * dt  # 更新速度
        if new_observation.vehicle_info['ego']['v_mps'] < -10/3.6:
            new_observation.vehicle_info['ego']['v_mps'] = -10/3.6

        new_observation.vehicle_info['ego']['acc_mpss'] = a  # 更新加速度
        return new_observation

    def _judge_gear(self, a, rot):
        # TODO 完成人字形节点判断，进而判断档位
        return 1
    
    def _get_slope_value(self,x, y):
        """
        根据车辆的位置 x, y 检索坡度值
        """
        return 0
    
    def _update_ego(self,unpacked_data,i,a,old_observation:Observation) -> Observation:
        "更新本车信息"
        new_observation = copy.copy(old_observation)
        new_observation.test_setting['t'] = float(
            old_observation.test_setting['t'] +
            old_observation.test_setting['dt']
            )
        new_observation.vehicle_info['ego']['x'] = unpacked_data[i*4+2]
        new_observation.vehicle_info['ego']['y'] = unpacked_data[i*4+3]
        # 更新航向角
        new_observation.vehicle_info['ego']['yaw_rad'] = unpacked_data[i*4+1]
        # 更新速度
        new_observation.vehicle_info['ego']['v_mps'] = unpacked_data[i*4+0]
        if new_observation.vehicle_info['ego']['v_mps'] < -10/3.6:
            new_observation.vehicle_info['ego']['v_mps'] = -10/3.6
        # 更新加速度
        new_observation.vehicle_info['ego']['acc_mpss'] = a
        return new_observation
    
    
    def _update_kinetics(self,env,action:tuple,old_observation:Observation,traj_future:Dict,observation_last:Observation,traj, client:Client, ha_path, slope=0, last_height=None, vis_cir_open=False) -> Observation:
        """
        使用动力学模型更新状态值
        """
        # 首先分别取出加速度和方向盘转角
        a,rot,gear = action
        dt = old_observation.test_setting['dt']
        #rot=1
        
        continue_simulation=0
        client.send_message(gear,a,rot,continue_simulation, slope)
        unpacked_data,step_count = client.receive_message()
        for i in range(step_count):
            # 更新状态
            new_observation=self._update_ego(unpacked_data,i,a,old_observation)
            new_observation,new_observation_all=self._update_other_vehicles(new_observation)
            new_observation_all = self._update_end_status(new_observation_all)
            #记录轨迹与可视化
            env.recorder.record(new_observation_all)
            env.visualizer.update(new_observation_all,traj_future,old_observation,traj,ha_path, last_height, vis_cir_open)

            if self.kinetics_mode == 'complex':
                if new_observation.test_setting['end'] != -1:
                    client.close_sockets()
                    break
            old_observation=new_observation

        if new_observation_all.test_setting['end'] == -1:
            #持续运行模型，直至收到下一个控制量
            continue_simulation=1
            client.send_message(gear, a, rot, continue_simulation, slope)
            
        return new_observation,new_observation_all

    def _update_other_vehicles(self,old_observation:Observation) -> Observation:
        """已知仿真时刻，把背景车信息更新至对应时刻(只返回30m内的车辆信息)"""
        # 删除除了ego之外的车辆观察值
        new_observation = copy.copy(old_observation)  # 复制一份旧观察值
        new_observation.vehicle_info = {}
        # 将本车信息添加回来
        new_observation.vehicle_info['ego'] = old_observation.vehicle_info['ego']
        new_observation_all = copy.deepcopy(new_observation)  # 包含所有车辆信息的观测值
        # 根据时间t,查询control_info,赋予新值
        t = old_observation.test_setting['t']
        t = str(np.around(t,2))  # t保留3位小数,与生成control_info时相吻合
        # 获取 ego 车辆的位置信息
        ego_x = old_observation.vehicle_info['ego']['x']
        ego_y = old_observation.vehicle_info['ego']['y']

        for vehi in self.control_info.vehicle_traj.items():
            id = vehi[0]  # 车辆id
            info = vehi[1]  # 车辆的轨迹信息
            if t in info.keys():
                new_observation_all.vehicle_info[id] = {}
                for key in ['x','y','yaw_rad','v_mps','yawrate_radps','acc_mpss']:
                    new_observation_all.vehicle_info[id][key] = info[t][key]
                new_observation_all.vehicle_info[id]['shape'] = info['shape']
                # 获取车辆在当前时刻的位置信息
                veh_x = info[t]['x']
                veh_y = info[t]['y']
                # 计算与 ego 车辆的距离
                distance = np.sqrt((veh_x - ego_x) ** 2 + (veh_y - ego_y) ** 2)
                #print("dis:",distance)
                if distance <= 40:
                    new_observation.vehicle_info[id] = {}
                    for key in ['x','y','yaw_rad','v_mps','yawrate_radps','acc_mpss']:
                        new_observation.vehicle_info[id][key] = info[t][key]
                    new_observation.vehicle_info[id]['shape'] = info['shape']

        return new_observation,new_observation_all

    def _update_end_status(self,observation:Observation) -> Observation:
        """计算T时刻,测试是否终止,更新observation.test_setting中的end值
            end=
                -1:回放测试正常进行; 
                1:回放测试运行完毕;
                2:ego车与其它车辆发生碰撞;
                3:ego车与道路边界发生碰撞(驶出道路边界);
                4:ego车到达目标区域               
        """
        status_list = [-1]

        # 检查主车与背景车是否发生碰撞
        if self._collision_detect(observation):
            status_list += [2]  # 添加状态
            print("###log### 主车与背景车发生碰撞\n")
            

        # 检查是否已到达场景终止时间max_t
        if observation.test_setting['t'] >= self.control_info.test_setting['max_t']:
            status_list += [1]
            print("###log### 已到达场景终止时间max_t\n")
            
    
        # 检查是否与道路边界碰撞
        local_x_range = observation.hdmaps['image_mask'].bitmap_info['bitmap_mask_PNG']['UTM_info']['local_x_range']
        local_y_range = observation.hdmaps['image_mask'].bitmap_info['bitmap_mask_PNG']['UTM_info']['local_y_range']
        collision_flag = self.collision_lookup.collision_detection(observation.vehicle_info['ego']['x']-local_x_range[0],
                                                        observation.vehicle_info['ego']['y']-local_y_range[0],
                                                        observation.vehicle_info['ego']['yaw_rad'],
                                                        observation.hdmaps['image_mask'].image_ndarray)
        if collision_flag == True:
            status_list += [3]
            print("###log### 主车与道路边界碰撞\n")
            
        # check target area
        if observation.test_setting["scenario_type"] == "intersection" or observation.test_setting["scenario_type"] == "unloading":
            if utils.is_inside_polygon(observation.vehicle_info['ego']['x'],observation.vehicle_info['ego']['y'],observation.test_setting['goal']):
                status_list += [4]
                print("###log### 主车已到达目标区域\n")
        elif observation.test_setting["scenario_type"] == "loading":
            if observation.test_setting["enter_loading_flag"] == False:
                if abs(observation.vehicle_info['ego']['v_mps']) < 1e-3 and \
                        utils.is_inside_polygon(observation.vehicle_info['ego']['x'], observation.vehicle_info['ego']['y'], observation.test_setting['goal']):
                    # status_list += [4]
                    print("###log### 主车已驶入装载区域，正在进行最终位姿调整\n")
                    observation.test_setting["enter_loading_flag"] = True
                    observation.test_setting['enter_loading_time'] = observation.test_setting['t']
            elif observation.test_setting["enter_loading_flag"] == True:
                adjusting_time = float(observation.test_setting['t'] - observation.test_setting["enter_loading_time"])  # 注意这里需要检查数据类型
                if adjusting_time > 5:
                    status_list += [4]
                    print("###log### 主车已驶入装载区域，且到达位姿调整时间上限\n")
        elif observation.test_setting["scenario_type"] == "mission":
            if utils.is_inside_polygon(observation.vehicle_info['ego']['x'],observation.vehicle_info['ego']['y'],observation.test_setting['goal']):
                status_list += [4]
                print("###log### 主车已到达目标区域\n")

        # 从所有status中取最大的那个作为end.
        observation.test_setting['end'] = max(status_list)
        return observation


    def _collision_detect(self,observation:Observation) -> bool:
        poly_zip = []
        self.vehicle_index = []  # 这里为了判断哪两辆车发生了碰撞,定义了列表用来存放车辆名称,其index与poly_zip中车辆的图形索引相对应
        # 当测试时间大于0.5秒时,遍历所有车辆,绘制对应的多边形.
        if observation.test_setting['t'] > 0.5:
            for index,vehi in observation.vehicle_info.items():
                self.vehicle_index += [index]
                poly_zip += [self._get_poly(vehi)]

        # 检测主车是否与背景车碰撞
        for a,b in combinations(poly_zip,2):
            if self.vehicle_index[poly_zip.index(a)] == 'ego' or self.vehicle_index[poly_zip.index(b)] == 'ego':
                if a.intersects(b):
                    return True
                else:
                    continue
        return False


    def _get_poly(self,vehicle:dict) -> Polygon:
        """根据车辆信息,通过shapely库绘制矩形.这是为了方便地使用shapely库判断场景中的车辆是否发生碰撞
        """
        # 提取车辆shape中的属性
        length = vehicle['shape']['length']
        width = vehicle['shape']['width']
        locationPoint2Head = vehicle['shape']['locationPoint2Head']
        locationPoint2Rear = vehicle['shape']['locationPoint2Rear']
    
        front_left_corner,front_right_corner,rear_left_corner,rear_right_corner = utils.calculate_vehicle_corners(
            length,width,locationPoint2Head,locationPoint2Rear,vehicle['x'],vehicle['y'],vehicle['yaw_rad'])

        # 通过车辆矩形的4个顶点,可以绘制出对应的长方形
        poly = Polygon([
            (front_left_corner[0],front_left_corner[1]),
            (front_right_corner[0],front_right_corner[1]),
            (rear_right_corner[0],rear_right_corner[1]),
            (rear_left_corner[0],rear_left_corner[1]),
            (front_left_corner[0],front_left_corner[1])
        ]).convex_hull
        return poly




class Controller():
    """控制车辆运行
    """
    def __init__(self) -> None:
        self.observation = Observation()
        self.parser = None
        self.control_info = None
        self.controller = None
        self.mode = 'replay'


    def init(self,scenario:dict, kinetics_mode:str, scene_type) -> Tuple:
        """初始化运行场景,给定初始时刻的观察值

        Parameters
        ----------
        input_dir :str
            测试输入文件所在位置
                回放测试:包含ScenariosResultes、TrajDataResultes、other目录,存放场景信息,轨迹片段集合、车辆配置信息等
                交互测试:
        mode :str
            指定测试模式
                回放测试:replay
                交互测试:interact
        Returns
        -------
        observation :Observation
            初始时刻的观察值信息,以Observation类的对象返回.
        """
        self.mode = scenario['test_settings']['mode']
        if self.mode == 'replay':
            self.parser = ReplayParser()
            self.controller = ReplayController(kinetics_mode)
            self.control_info = self.parser.parse(scenario, scene_type)
            observation,self.observation = self.controller.init(self.control_info)
            self.traj = self.control_info.vehicle_traj
        return observation,self.observation,self.traj


    def step(self,env,action, client:None,traj_future:Dict,observation_last:Observation,traj=200, ha_path=None, slope=0, last_height=None, vis_cir_open=False):
        """
        根据控制量更新状态值，只返回限制范围内观测值
        observation_dis为限制感知范围得观测值
        self.observation包含全部观测信息，用于可视化
        """
        observation_dis,self.observation = self.controller.step(env,action,self.observation,traj_future,observation_last,traj,ha_path, client, slope, last_height, vis_cir_open)
        return observation_dis




if __name__ == "__main__":
    input_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                             '/home/czf/project_czf/20231010_onsite_mine/devkit/inputs'))
    scenes_file = os.path.join(input_dir,'ScenariosResultes','Scenario-jiangtong_intersection_1_1_2.json')

    with open(scenes_file,'r') as f:
        jsondata = json.load(f)
            
    pass



# -*- coding:utf-8 -*-


"""A Python implemntation of a kd-tree

This package provides a simple implementation of a kd-tree in Python.
https://en.wikipedia.org/wiki/K-d_tree
"""

from __future__ import print_function

import heapq
import itertools
import operator
import math
from collections import deque
from functools import wraps

__author__ = "Stefan Kögl <stefan@skoegl.net>"
__version__ = "0.16"
__website__ = "https://github.com/stefankoegl/kdtree"
__license__ = "ISC license"


class Node(object):
    """A Node in a kd-tree

    A tree is represented by its root node,and every node represents
    its subtree"""

    def __init__(self, data=None, left=None, right=None):
        self.data = data
        self.left = left
        self.right = right

    @property
    def is_leaf(self):
        """Returns True if a Node has no subnodes

        >>> Node().is_leaf
        True

        >>> Node( 1,left=Node(2) ).is_leaf
        False
        """
        return (not self.data) or (all(not bool(c) for c, p in self.children))

    def preorder(self):
        """iterator for nodes:root,left,right"""

        if not self:
            return

        yield self

        if self.left:
            for x in self.left.preorder():
                yield x

        if self.right:
            for x in self.right.preorder():
                yield x

    def inorder(self):
        """iterator for nodes:left,root,right"""

        if not self:
            return

        if self.left:
            for x in self.left.inorder():
                yield x

        yield self

        if self.right:
            for x in self.right.inorder():
                yield x

    def postorder(self):
        """iterator for nodes:left,right,root"""

        if not self:
            return

        if self.left:
            for x in self.left.postorder():
                yield x

        if self.right:
            for x in self.right.postorder():
                yield x

        yield self

    @property
    def children(self):
        """
        Returns an iterator for the non-empty children of the Node

        The children are returned as (Node,pos) tuples where pos is 0 for the
        left subnode and 1 for the right.

        >>> len(list(create(dimensions=2).children))
        0

        >>> len(list(create([ (1,2) ]).children))
        0

        >>> len(list(create([ (2,2),(2,1),(2,3) ]).children))
        2
        """

        if self.left and self.left.data is not None:
            yield self.left, 0
        if self.right and self.right.data is not None:
            yield self.right, 1

    def set_child(self, index, child):
        """Sets one of the node's children

        index 0 refers to the left,1 to the right child"""

        if index == 0:
            self.left = child
        else:
            self.right = child

    def height(self):
        """
        Returns height of the (sub)tree,without considering
        empty leaf-nodes

        >>> create(dimensions=2).height()
        0

        >>> create([ (1,2) ]).height()
        1

        >>> create([ (1,2),(2,3) ]).height()
        2
        """

        min_height = int(bool(self))
        return max([min_height] + [c.height() + 1 for c, p in self.children])

    def get_child_pos(self, child):
        """Returns the position if the given child

        If the given node is the left child,0 is returned. If its the right
        child,1 is returned. Otherwise None"""

        for c, pos in self.children:
            if child == c:
                return pos

    def __repr__(self):
        return "<%(cls)s - %(data)s>" % dict(cls=self.__class__.__name__, data=repr(self.data))

    def __nonzero__(self):
        return self.data is not None

    __bool__ = __nonzero__

    def __eq__(self, other):
        if isinstance(other, tuple):
            return self.data == other
        else:
            return self.data == other.data

    def __hash__(self):
        return id(self)


def require_axis(f):
    """Check if the object of the function has axis and sel_axis members"""

    @wraps(f)
    def _wrapper(self, *args, **kwargs):
        if None in (self.axis, self.sel_axis):
            raise ValueError(
                "%(func_name) requires the node %(node)s " "to have an axis and a sel_axis function" % dict(func_name=f.__name__, node=repr(self))
            )

        return f(self, *args, **kwargs)

    return _wrapper


class KDNode(Node):
    """A Node that contains kd-tree specific data and methods"""

    def __init__(self, data=None, left=None, right=None, axis=None, sel_axis=None, dimensions=None):
        """Creates a new node for a kd-tree

        If the node will be used within a tree,the axis and the sel_axis
        function should be supplied.

        sel_axis(axis) is used when creating subnodes of the current node. It
        receives the axis of the parent node and returns the axis of the child
        node."""
        super(KDNode, self).__init__(data, left, right)
        self.axis = axis
        self.sel_axis = sel_axis
        self.dimensions = dimensions

    @require_axis
    def add(self, point):
        """
        Adds a point to the current node or iteratively
        descends to one of its children.

        Users should call add() only to the topmost tree.
        """

        current = self
        while True:
            check_dimensionality([point], dimensions=current.dimensions)

            # Adding has hit an empty leaf-node,add here
            if current.data is None:
                current.data = point
                return current

            # split on self.axis,recurse either left or right
            if point[current.axis] < current.data[current.axis]:
                if current.left is None:
                    current.left = current.create_subnode(point)
                    return current.left
                else:
                    current = current.left
            else:
                if current.right is None:
                    current.right = current.create_subnode(point)
                    return current.right
                else:
                    current = current.right

    @require_axis
    def create_subnode(self, data):
        """Creates a subnode for the current node"""

        return self.__class__(data, axis=self.sel_axis(self.axis), sel_axis=self.sel_axis, dimensions=self.dimensions)

    @require_axis
    def find_replacement(self):
        """Finds a replacement for the current node

        The replacement is returned as a
        (replacement-node,replacements-parent-node) tuple"""

        if self.right:
            child, parent = self.right.extreme_child(min, self.axis)
        else:
            child, parent = self.left.extreme_child(max, self.axis)

        return (child, parent if parent is not None else self)

    def should_remove(self, point, node):
        """checks if self's point (and maybe identity) matches"""
        if not self.data == point:
            return False

        return (node is None) or (node is self)

    @require_axis
    def remove(self, point, node=None):
        """Removes the node with the given point from the tree

        Returns the new root node of the (sub)tree.

        If there are multiple points matching "point",only one is removed. The
        optional "node" parameter is used for checking the identity,once the
        removeal candidate is decided."""

        # Recursion has reached an empty leaf node,nothing here to delete
        if not self:
            return

        # Recursion has reached the node to be deleted
        if self.should_remove(point, node):
            return self._remove(point)

        # Remove direct subnode
        if self.left and self.left.should_remove(point, node):
            self.left = self.left._remove(point)

        elif self.right and self.right.should_remove(point, node):
            self.right = self.right._remove(point)

        # Recurse to subtrees
        if point[self.axis] <= self.data[self.axis]:
            if self.left:
                self.left = self.left.remove(point, node)

        if point[self.axis] >= self.data[self.axis]:
            if self.right:
                self.right = self.right.remove(point, node)

        return self

    @require_axis
    def _remove(self, point):
        # we have reached the node to be deleted here

        # deleting a leaf node is trivial
        if self.is_leaf:
            self.data = None
            return self

        # we have to delete a non-leaf node here

        # find a replacement for the node (will be the new subtree-root)
        root, max_p = self.find_replacement()

        # self and root swap positions
        tmp_l, tmp_r = self.left, self.right
        self.left, self.right = root.left, root.right
        root.left, root.right = tmp_l if tmp_l is not root else self, tmp_r if tmp_r is not root else self
        self.axis, root.axis = root.axis, self.axis

        # Special-case if we have not chosen a direct child as the replacement
        if max_p is not self:
            pos = max_p.get_child_pos(root)
            max_p.set_child(pos, self)
            max_p.remove(point, self)

        else:
            root.remove(point, self)

        return root

    @property
    def is_balanced(self):
        """Returns True if the (sub)tree is balanced

        The tree is balanced if the heights of both subtrees differ at most by
        1"""

        left_height = self.left.height() if self.left else 0
        right_height = self.right.height() if self.right else 0

        if abs(left_height - right_height) > 1:
            return False

        return all(c.is_balanced for c, _ in self.children)

    def rebalance(self):
        """
        Returns the (possibly new) root of the rebalanced tree
        """

        return create([x.data for x in self.inorder()])

    def axis_dist(self, point, axis):
        """
        Squared distance at the given axis between
        the current Node and the given point
        """
        return math.pow(self.data[axis] - point[axis], 2)

    def dist(self, point):
        """
        Squared distance between the current Node
        and the given point
        """
        r = range(self.dimensions)
        return sum([self.axis_dist(point, i) for i in r])

    def search_knn(self, point, k, dist=None):
        """Return the k nearest neighbors of point and their distances

        point must be an actual point,not a node.

        k is the number of results to return. The actual results can be less
        (if there aren't more nodes to return) or more in case of equal
        distances.

        dist is a distance function,expecting two points and returning a
        distance value. Distance values can be any comparable type.

        The result is an ordered list of (node,distance) tuples.
        """

        if k < 1:
            raise ValueError("k must be greater than 0.")

        if dist is None:
            get_dist = lambda n: n.dist(point)
        else:
            get_dist = lambda n: dist(n.data, point)

        results = []

        self._search_node(point, k, results, get_dist, itertools.count())

        # We sort the final result by the distance in the tuple
        # (<KdNode>,distance).
        return [(node, -d) for d, _, node in sorted(results, reverse=True)]

    def _search_node(self, point, k, results, get_dist, counter):
        if not self:
            return

        nodeDist = get_dist(self)

        # Add current node to the priority queue if it closer than
        # at least one point in the queue.
        #
        # If the heap is at its capacity,we need to check if the
        # current node is closer than the current farthest node,and if
        # so,replace it.
        item = (-nodeDist, next(counter), self)
        if len(results) >= k:
            if -nodeDist > results[0][0]:
                heapq.heapreplace(results, item)
        else:
            heapq.heappush(results, item)
        # get the splitting plane
        split_plane = self.data[self.axis]
        # get the squared distance between the point and the splitting plane
        # (squared since all distances are squared).
        plane_dist = point[self.axis] - split_plane
        plane_dist2 = plane_dist * plane_dist

        # Search the side of the splitting plane that the point is in
        if point[self.axis] < split_plane:
            if self.left is not None:
                self.left._search_node(point, k, results, get_dist, counter)
        else:
            if self.right is not None:
                self.right._search_node(point, k, results, get_dist, counter)

        # Search the other side of the splitting plane if it may contain
        # points closer than the farthest point in the current results.
        if -plane_dist2 > results[0][0] or len(results) < k:
            if point[self.axis] < self.data[self.axis]:
                if self.right is not None:
                    self.right._search_node(point, k, results, get_dist, counter)
            else:
                if self.left is not None:
                    self.left._search_node(point, k, results, get_dist, counter)

    @require_axis
    def search_nn(self, point, dist=None):
        """
        Search the nearest node of the given point

        point must be an actual point,not a node. The nearest node to the
        point is returned. If a location of an actual node is used,the Node
        with this location will be returned (not its neighbor).

        dist is a distance function,expecting two points and returning a
        distance value. Distance values can be any comparable type.

        The result is a (node,distance) tuple.
        """

        return next(iter(self.search_knn(point, 1, dist)), None)

    def _search_nn_dist(self, point, dist, results, get_dist):
        if not self:
            return

        nodeDist = get_dist(self)

        if nodeDist < dist:
            results.append(self.data)

        # get the splitting plane
        split_plane = self.data[self.axis]

        # Search the side of the splitting plane that the point is in
        if point[self.axis] <= split_plane + dist:
            if self.left is not None:
                self.left._search_nn_dist(point, dist, results, get_dist)
        if point[self.axis] >= split_plane - dist:
            if self.right is not None:
                self.right._search_nn_dist(point, dist, results, get_dist)

    @require_axis
    def search_nn_dist(self, point, distance, best=None):
        """
        Search the n nearest nodes of the given point which are within given
        distance

        point must be a location,not a node. A list containing the n nearest
        nodes to the point within the distance will be returned.
        """

        results = []
        get_dist = lambda n: n.dist(point)

        self._search_nn_dist(point, distance, results, get_dist)
        return results

    @require_axis
    def is_valid(self):
        """Checks recursively if the tree is valid

        It is valid if each node splits correctly"""

        if not self:
            return True

        if self.left and self.data[self.axis] < self.left.data[self.axis]:
            return False

        if self.right and self.data[self.axis] > self.right.data[self.axis]:
            return False

        return all(c.is_valid() for c, _ in self.children) or self.is_leaf

    def extreme_child(self, sel_func, axis):
        """Returns a child of the subtree and its parent

        The child is selected by sel_func which is either min or max
        (or a different function with similar semantics)."""

        max_key = lambda child_parent: child_parent[0].data[axis]

        # we don't know our parent,so we include None
        me = [(self, None)] if self else []

        child_max = [c.extreme_child(sel_func, axis) for c, _ in self.children]
        # insert self for unknown parents
        child_max = [(c, p if p is not None else self) for c, p in child_max]

        candidates = me + child_max

        if not candidates:
            return None, None

        return sel_func(candidates, key=max_key)


def create(point_list=None, dimensions=None, axis=0, sel_axis=None):
    """Creates a kd-tree from a list of points

    All points in the list must be of the same dimensionality.

    If no point_list is given,an empty tree is created. The number of
    dimensions has to be given instead.

    If both a point_list and dimensions are given,the numbers must agree.

    Axis is the axis on which the root-node should split.

    sel_axis(axis) is used when creating subnodes of a node. It receives the
    axis of the parent node and returns the axis of the child node."""

    if not point_list and not dimensions:
        raise ValueError("either point_list or dimensions must be provided")

    elif point_list:
        dimensions = check_dimensionality(point_list, dimensions)

    # by default cycle through the axis
    sel_axis = sel_axis or (lambda prev_axis: (prev_axis + 1) % dimensions)

    if not point_list:
        return KDNode(sel_axis=sel_axis, axis=axis, dimensions=dimensions)

    # Sort point list and choose median as pivot element
    point_list = list(point_list)
    point_list.sort(key=lambda point: point[axis])
    median = len(point_list) // 2

    loc = point_list[median]
    left = create(point_list[:median], dimensions, sel_axis(axis))
    right = create(point_list[median + 1 :], dimensions, sel_axis(axis))
    return KDNode(loc, left, right, axis=axis, sel_axis=sel_axis, dimensions=dimensions)


def check_dimensionality(point_list, dimensions=None):
    dimensions = dimensions or len(point_list[0])
    for p in point_list:
        if len(p) != dimensions:
            raise ValueError("All Points in the point_list must have the same dimensionality")

    return dimensions


def level_order(tree, include_all=False):
    """Returns an iterator over the tree in level-order

    If include_all is set to True,empty parts of the tree are filled
    with dummy entries and the iterator becomes infinite."""

    q = deque()
    q.append(tree)
    while q:
        node = q.popleft()
        yield node

        if include_all or node.left:
            q.append(node.left or node.__class__())

        if include_all or node.right:
            q.append(node.right or node.__class__())


def visualize(tree, max_level=100, node_width=10, left_padding=5):
    """Prints the tree to stdout"""

    height = min(max_level, tree.height() - 1)
    max_width = pow(2, height)

    per_level = 1
    in_level = 0
    level = 0

    for node in level_order(tree, include_all=True):

        if in_level == 0:
            print()
            print()
            print(" " * left_padding, end=" ")

        width = int(max_width * node_width / per_level)

        node_str = (str(node.data) if node else "").center(width)
        print(node_str, end=" ")

        in_level += 1

        if in_level == per_level:
            in_level = 0
            per_level *= 2
            level += 1

        if level > height:
            break

    print()
    print()



import sys
import os
from pathlib import Path as PathlibPath
from typing import Dict,List,Tuple,Optional,Union
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),'..')))

# 第三方库
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import Path as MatplotlibPath
import matplotlib.patches as patches
from matplotlib.patches import Rectangle,Arrow
from matplotlib.axes import Axes

# 自定义模块
from dynamic_scenes.observation import Observation
from map_expansion.map_api import TgScenesMap,TgScenesMapExplorer
from map_expansion.bit_map import BitMap



def check_dir(target_dir:str) -> None:
    """
    Check and create the directory if it doesn't exist.

    Parameters:
    - target_dir (str):The directory path to be checked.
    """
    if not os.path.exists(target_dir):
        os.makedirs(target_dir,exist_ok=True)  # If directory already exists,don't raise an error.





class Visualizer:
    COLOR_MAP = {
        'background_vehicle':'#6495ED', # "cornflowerblue" 矢车菊蓝;矢车菊的蓝色
        'static_obstacle':'#4682B4',  # 静态障碍物，钢青色
        'ego_vehicle':'#008000',  # 深绿色，主车
        'warning_signal':'#e60000',  # 红色，警告标志
        'traj_prediction_line':'#4de680',  # 预测轨迹，绿松石色
        'traj_prediction_point':'#127436',  # 预测轨迹点，深绿色
        'traj_true_line':'#e60000',  # 红色，真实轨迹
        'traj_true_point':'#9B0000',  # 深红色，真实轨迹点
        'traj_true_history':'#eb984e',  # 历史轨迹，橙棕色
        'target_box_facecolor':'#FF6347'}  # 目标区域，橙红色
    X_MARGIN = 8 # 类常量
    Y_MARGIN = 8

    def __init__(self):
        self.flag_visilize = True # 是\否进行可视化绘图
        self.flag_hdmaps_visilized = False # hdmaps已经被绘图(仅绘制一次即可)标志位
        self.flag_visilize_ref_path = True #是否可视化参考路径
        self.flag_visilize_prediction = False #是否可视化轨迹预测结果
        self.flag_save_fig_whitout_show = False # 绘图保存,但是不显示
        
    def battle_init(self,observation:"Observation",flag_visilize:bool = False,flag_save_fig_whitout_show:bool = False, img_save_path='') -> None:
         # 测试信息
        self.scene_type = "B"
        self.scenario_name = observation.test_setting['scenario_name']
        self.scenario_type = observation.test_setting['scenario_type']
        self.save_path_img = os.path.join(img_save_path , f"image_{self.scenario_name}")
        if not os.path.exists(self.save_path_img):
            os.makedirs(self.save_path_img)
            
        image_path = os.path.join(self.save_path_img , "images")
        # 检查文件夹是否存在
        if os.path.exists(image_path) and os.path.isdir(image_path):
            # 如果文件夹存在，清空其中的所有文件
            file_list = os.listdir(image_path)
            for file in file_list:
                file_path = os.path.join(image_path, file)
                if os.path.isfile(file_path):
                    os.remove(file_path)
        else:
            # 如果文件夹不存在，可以选择创建它
            os.makedirs(image_path)
            print(f"文件夹不存在，已创建：{self.save_path_img}及其images子文件夹")
        
        # 可视化设置
        self.flag_visilize = flag_visilize
        self.flag_save_fig_whitout_show = flag_save_fig_whitout_show
        
        # 初始化画布
        if self.flag_visilize:
            if self.flag_save_fig_whitout_show == True:
                plt.ioff()  # 关闭交互模式，没有plt.show不显示（即不显示可视化）
            else:
                plt.ion()  # 交互模式，随时显示，不需要plt.show
            if (observation.test_setting['x_max']-observation.test_setting['x_min']):# 防止除0错误
                temp_fig_y_coeff = (observation.test_setting['y_max']-observation.test_setting['y_min'])/(observation.test_setting['x_max']-observation.test_setting['x_min'])
            else:
                temp_fig_y_coeff = 1
            self.fig = plt.figure(figsize=(10.0, 10.0*temp_fig_y_coeff))
            # self.fig = plt.figure(figsize=[5.0,5.0])
            self.axbg = self.fig.add_subplot()  # 多个x轴,共用y轴
            self.axveh = self.axbg.twiny()
            self.axTraj = self.axbg.twiny()
        else:
            plt.ioff()
            return

        self.flag_hdmaps_visilized = False # 本次绘制地图
        if observation.hdmaps:
            self.x_target = list(observation.test_setting['goal']['x'])
            self.y_target = list(observation.test_setting['goal']['y'])
            self.heading = observation.test_setting['goal']['heading']
            # 获取绘图范围
            bitmap_info = observation.hdmaps['image_mask'].bitmap_local_info
            self.x_min = bitmap_info['utm_local_range'][0]
            self.x_max = bitmap_info['utm_local_range'][2]
            self.y_min = bitmap_info['utm_local_range'][1]
            self.y_max = bitmap_info['utm_local_range'][3]
            self.axbg.set_xlim(self.x_min - self.X_MARGIN,self.x_max + self.X_MARGIN)
        
        # 更新观测值 and 绘图刷新(第一次进入)
        self.update(observation,traj_future=100,observation_last=100,traj=100)
    def init(self,observation:"Observation",flag_visilize:bool = False,flag_save_fig_whitout_show:bool = False, img_save_path='') -> None:
        """绘图,第一次进入,初始化.

        Args:
            observation (Observation): 当前时刻对动态场景、静态地图的观测结果.
            flag_visilize (bool, optional): 是\否进行可视化绘图. Defaults to False.
        """
        # 测试信息
        self.scene_type = "not_battlefied"
        self.scenario_name = observation.test_setting['scenario_name']
        self.scenario_type = observation.test_setting['scenario_type']
        self.save_path_img = os.path.join(img_save_path , f"image_{self.scenario_name}")
        if not os.path.exists(self.save_path_img):
            os.makedirs(self.save_path_img)
            
        image_path = os.path.join(self.save_path_img , "images")
        # 检查文件夹是否存在
        if os.path.exists(image_path) and os.path.isdir(image_path):
            # 如果文件夹存在，清空其中的所有文件
            file_list = os.listdir(image_path)
            for file in file_list:
                file_path = os.path.join(image_path, file)
                if os.path.isfile(file_path):
                    os.remove(file_path)
        else:
            # 如果文件夹不存在，可以选择创建它
            os.makedirs(image_path)
            print(f"文件夹不存在，已创建：{self.save_path_img}及其images子文件夹")
        
        # 可视化设置
        self.flag_visilize = flag_visilize
        self.flag_save_fig_whitout_show = flag_save_fig_whitout_show
        
        # 初始化画布
        if self.flag_visilize:
            if self.flag_save_fig_whitout_show == True:
                plt.ioff()  # 关闭交互模式，没有plt.show不显示（即不显示可视化）
            else:
                plt.ion()  # 交互模式，随时显示，不需要plt.show
            self.fig = plt.figure(figsize=(10.0,10.0))
            # self.fig = plt.figure(figsize=[5.0,5.0])
            self.axbg = self.fig.add_subplot()  # 多个x轴,共用y轴
            self.axveh = self.axbg.twiny()
            self.axTraj = self.axbg.twiny()
        else:
            plt.ioff()
            return

        self.flag_hdmaps_visilized = False # 本次绘制地图
        if observation.hdmaps:
            self.x_target = list(observation.test_setting['goal']['x'])
            self.y_target = list(observation.test_setting['goal']['y'])
            self.heading = observation.test_setting['goal']['head']
            # 获取绘图范围
            bitmap_info = observation.hdmaps['image_mask'].bitmap_local_info
            self.x_min = bitmap_info['utm_local_range'][0]
            self.x_max = bitmap_info['utm_local_range'][2]
            self.y_min = bitmap_info['utm_local_range'][1]
            self.y_max = bitmap_info['utm_local_range'][3]
            self.axbg.set_xlim(self.x_min - self.X_MARGIN,self.x_max + self.X_MARGIN)
        
        # 更新观测值 and 绘图刷新(第一次进入)
        self.update(observation,traj_future=100,observation_last=100,traj=100)

    def _plot_battlefield_hdmaps(self,observation:Observation,) -> None:
        if not observation.hdmaps:
            return
        # plot mask图,plot 道路片段、交叉口多边形划分;
        my_patch = (self.x_min,self.y_min ,self.x_max,self.y_max)  #  (x_min,y_min,x_max,y_max).
        # layer_names = ['intersection','road']
        layer_names =  ['road','intersection','loading_area','unloading_area','road_block']
        self._plot_battlefield_hdmaps_render_map_patch(
                                           box_coords=my_patch,layer_names=layer_names,alpha=0.7,  # 坐标范围，需要绘制的层，渗透率
                                           render_egoposes_range=False,render_legend=False,
                                           bitmap=observation.hdmaps['image_mask'],  # Bitmap类
                                           ax= self.axbg)
    def _plot_battlefield_target_area(self,observation:Observation):   
        self.x_target = list(observation.test_setting['goal']['x'])
        self.y_target = list(observation.test_setting['goal']['y'])
        self.heading = observation.test_setting['goal']['heading']
        if self.x_target and self.y_target:
            x,y = self.x_target,self.y_target
            codes_box = [MatplotlibPath.MOVETO] + [MatplotlibPath.LINETO] * 3 + [MatplotlibPath.CLOSEPOLY] 
            vertices_box = [(self.x_target[0],self.y_target[0]),
                            (self.x_target[1],self.y_target[1]),
                            (self.x_target[2],self.y_target[2]),
                            (self.x_target[3],self.y_target[3]),
                            (self.x_target[0],self.y_target[0])] # 4 points of polygon
            path_box = MatplotlibPath(vertices_box,codes_box)  # 定义对应Path
            pathpatch_box = patches.PathPatch(path_box,
                                              facecolor=self.COLOR_MAP['target_box_facecolor'],
                                              edgecolor='orangered',
                                              zorder=2,
                                              alpha=0.7
                                             )
            self.axbg.add_patch(pathpatch_box)
            center_x = observation.test_setting["point_goal"][0]
            center_y = observation.test_setting["point_goal"][1]
            node_ux = 5*np.cos(self.heading[0])
            node_vy = 5*np.sin(self.heading[0])
            self.axbg.arrow(center_x, center_y, node_ux, node_vy, color='red', alpha=1, head_width=2.0, head_length=2.0)
        else:
            x = observation.test_setting["point_goal"][0]
            y = observation.test_setting["point_goal"][1]
            self.axbg.plot(x, y, 'o', markersize=2, color='red')

    def update(self,observation:Observation,traj_future,observation_last,traj, ha_path=None, last_height=None, vis_cir_open=False) -> None:
        """更新观测值 and 绘图刷新.

        Args:
            observation (Observation): 当前时刻对动态场景、静态地图的观测结果.
            traj_future (_type_): 背景车预测的轨迹.
            observation_last (_type_): 上一刻 动态场景、静态地图的观测结果.
            traj (_type_):该场景背景车辆全部轨迹(ground truth).
        """
        # --------结束条件如果不画图, 直接退出 --------
        if not self.flag_visilize:
            return
        self.axveh.cla()  # 清除当前子图中的所有绘图内容
        self.axTraj.cla()  # 清除当前子图中的所有绘图内容;

        # -------- 绘制道路 \ 绘制车辆 \  --------
        if not self.flag_hdmaps_visilized:
            if self.scene_type == "B":
                self._plot_battlefield_hdmaps(observation)
                self._plot_battlefield_target_area(observation)
            else:
                self._plot_hdmaps(observation)  # 绘制位图、语义地图的road、intersection、loading_area、unloading_area、dubins_pose以及reference_path
                self._plot_target_area(observation)  # 绘制目标区域(loading场景加载heading指向)
            self.flag_hdmaps_visilized = True  # 防止绘制多次
        self._plot_vehicles(observation)
        self._plot_hybrid_astar_path(ha_path)  # 绘制自车hybrid_astar的路径
        
        # --------若发生碰撞,则绘制碰撞警告标志--------
        if observation.test_setting['end'] in [2,3]:
            self._plot_warning_signal(observation)

        # --------绘制高度矩阵相关信息--------
        if last_height is not None:
            for i in range(len(last_height)):
                for j in range(len(last_height[i])):
                    x = int(observation.vehicle_info['ego']['x'])+0.5 - 20 + i
                    y = int(observation.vehicle_info['ego']['y'])+0.5 - 20 + j
                    if last_height[i][j] == -1:
                        plt.plot(x, y, marker='o', markersize=3, linestyle='None', color='cyan')

        # --------绘制观测范围--------
        if vis_cir_open == True:
            center = (observation.vehicle_info['ego']['x'], observation.vehicle_info['ego']['y'])
            radius = 28.2
            circle = patches.Circle(center, radius, 
                        edgecolor='cyan',    # 边缘颜色为青色
                        facecolor='cyan',    # 填充颜色为青色
                        linewidth=2,         # 边缘宽度
                        alpha=0.2)  # 透明度
            plt.gca().add_patch(circle)

        
        # --------绘制车辆轨迹:预测轨迹与真实轨迹--------
        if (traj_future != 100 and  traj != 100):#具备可视化 预测轨迹\真实轨迹 的条件
            if traj==200:
                self._plot_traj_future_path(traj_future)  # 绘制背景车预测的轨迹
                # print("test, 关闭全局背景车轨迹成功，启用自定义可视化预测轨迹功能")
            if self.flag_visilize_prediction == True:
                self._plot_vehicles_traj(observation,traj_future,observation_last,traj)

        # --------画布属性设置--------
        # plt.rcParams['font.family'] = 'Times New Roman'  # 若需显示汉字 SimHei 黑体, STsong 华文宋体还有 font.style  font.size等
        plt.rcParams['axes.unicode_minus'] = False
        if observation.hdmaps:# 有地图文件,采用主车中心视角
            plt.ylim(self.y_min - self.Y_MARGIN,self.y_max + self.Y_MARGIN)   # 指定y坐标的显示范围,防止车辆畸变,y轴范围与x轴一致
            self.axbg.set_xlim(self.x_min - self.X_MARGIN,self.x_max + self.X_MARGIN)  
            self.axveh.set_xlim(self.x_min - self.X_MARGIN,self.x_max + self.X_MARGIN)
            self.axTraj.set_xlim(self.x_min - self.X_MARGIN,self.x_max + self.X_MARGIN)  
            
        else: # 无地图文件的(如加速测试),采用主车中心视角
            plt.gca().set_aspect('equal')
            x_center = observation.vehicle_info['ego']['x']
            plt.xlim(x_center - 20,x_center + 200)
            plt.ylim(-80,80)

        # -------- GIF显示各类信息 --------
        # 显示测试相关信息
        ts_text = "Time_stamp:" + str(round(observation.test_setting['t'],4))
        name_text = 'Test_scenario:' + str(self.scenario_name)
        type_text = 'Task_type:' + str(self.scenario_type)
        self.axveh.text(0.02,1.05,name_text,transform=self.axveh.transAxes,fontdict={'size':'10','color':'black'})  # axveh.transAxes表示坐标比例系
        self.axveh.text(0.02,1.08,type_text,transform=self.axveh.transAxes,fontdict={'size':'10','color':'black'})
        self.axveh.text(0.02,1.11,ts_text,transform=self.axveh.transAxes,fontdict={'size':'10','color':'black'})
        
        #  碰撞信息显示 
        if observation.test_setting['end'] in [2,3]:
            collision_text = 'A collision has occurred.'
            self.axveh.text(0.02,0.85,collision_text,transform=self.axveh.transAxes,fontdict={'size':'18','color':'red'})
        
        #  显示所有车辆运行信息 
        colLabels = list(observation.vehicle_info.keys())
        rowLabels = ['v (m/s)','a (m/s2)'] 
        v = np.array([round(observation.vehicle_info[key]['v_mps'],4) for key in colLabels]).reshape(1,-1)
        a = np.array([round(observation.vehicle_info[key]['acc_mpss'],4) for key in colLabels]).reshape(1,-1)
        cellTexts = np.vstack((v,a))
        info_table = self.axveh.table(cellText=cellTexts,colLabels=colLabels,rowLabels=rowLabels,
                                      rowLoc='center',colLoc='center',cellLoc='center',loc='bottom')
        info_table.auto_set_font_size(False)
        info_table.set_fontsize(10)

        # --------刷新当前帧画布--------
        plt.subplots_adjust()
        step_sum = int(round(observation.test_setting['max_t'] / observation.test_setting['dt']))
        if self.flag_save_fig_whitout_show:
            # 想要把预测结果保存为gif,先把png保存下来.
            step = int(round(observation.test_setting['t'] / observation.test_setting['dt'])) 
            png_dir = os.path.join(self.save_path_img,'images')
            check_dir(png_dir)
            plt.savefig(os.path.join(png_dir,f"image_{observation.test_setting['scenario_name']}_{step}.png"))

            # print(f"###保存单帧图片### 第{step}帧/最大帧数{step_sum}")
        else:
            plt.pause(1e-7)
            plt.show()
        
        # -------- 如果测试结束,则结束绘图,关闭绘图模块--------
        if observation.test_setting['end'] != -1:
            plt.ioff()
            plt.close()
            return


    def _plot_vehicles_traj(self,observation:Observation,traj_future:Dict,observation_last:Observation,traj:Dict) -> None:
        """绘制车辆预测轨迹及真实轨迹.

        Args:
            observation (Observation): 当前时刻对动态场景、静态地图的观测结果.
            traj_future (_type_): 背景车预测的轨迹.
            observation_last (_type_): 上一刻 动态场景、静态地图的观测结果.
            traj (_type_):该场景背景车辆全部轨迹(ground truth).
        """
         
        for id_vehi,traj_predict in traj_future.items():
            # Extract true trajectory
            traj_true = traj.get(id_vehi,{}).copy()
            
            # Remove unwanted keys from true trajectory
            traj_true.pop('shape',None)
            traj_true.pop(-1,None)  # 没用但不会报错
            
            # Ensure the vehicle ID exists in the true trajectory
            if not traj_true:
                raise ValueError(f"Error:No vehicle with ID '{id_vehi}' found in the true trajectory.")
            
            # Plot individual vehicle trajectory
            self._plot_single_vehicle_traj(observation,id_vehi,traj_true,traj_predict)

  
    def _plot_single_vehicle_traj(self,observation:Observation,id_vehi:int,traj_true:Dict,traj_predict:Dict)-> None:
        """ 利用 matplotlib 绘制单个车的轨迹.
        """
        # 1) 从字典中提取x y轨迹点的列表
        # 1.1) 计算key值列表
        # 获取字典的键并转换为浮点数
        keys_as_floats = [float(key) for key in traj_true.keys()] 
        max_time_of_id_vehi = max(keys_as_floats)

        dt = round(float(observation.test_setting['dt']),2) # 时间步长
        now_time = round(float(observation.test_setting['t'])-dt,2)   # 当前时间
        end_time = float(observation.test_setting['max_t'])                
        delta_t_1 = end_time - now_time
        delta_t_2 = max_time_of_id_vehi - now_time
        delta_t = min(delta_t_1,delta_t_2)
        
        if delta_t < 5.0 :
            numOfTrajPoint_ = int(delta_t / dt)
            float_list,keys_to_extract_line= self._generate_float_list(now_time,dt,numOfTrajPoint_)  # 返回时间戳列表
            float_list_1s,keys_to_extract_line_1s= self._generate_float_list_2(now_time,1.0,end_time)
        else:
            numOfTrajPoint_ = int(5.0/ dt) # 5秒*10hz;
            float_list,keys_to_extract_line= self._generate_float_list(now_time,dt,numOfTrajPoint_)
            float_list_1s,keys_to_extract_line_1s = self._generate_float_list_2(now_time,1.0,now_time+5.0)       

        # 1.2) 提取 5s的所有轨迹点 + 1 2 3 4 5秒处的轨迹点
        xlinelist_true,ylinelist_true =  self._find_xylist_from_vehi_traj(keys_to_extract_line,traj_true)
        xpointlist_true,ypointlist_true = self._find_xylist_from_vehi_traj(keys_to_extract_line_1s,traj_true)
        self._plot_traj_line_points(xlinelist=xlinelist_true,
                                    ylinelist=ylinelist_true,
                                    xpointlist=xpointlist_true,
                                    ypointlist=ypointlist_true,
                                    color_line=self.COLOR_MAP['traj_true_line'],
                                    line_wide=2.0,
                                    color_point=self.COLOR_MAP['traj_true_point'],
                                    size_point=4.0)
        # 2.2) plot 预测的轨迹
        xlinelist_predict,ylinelist_predict =  self._find_xylist_from_vehi_traj(keys_to_extract_line,traj_predict)
        xpointlist_predict,ypointlist_predict = self._find_xylist_from_vehi_traj(keys_to_extract_line_1s,traj_predict)
        self._plot_traj_line_points(xlinelist=xlinelist_predict,
                                    ylinelist=ylinelist_predict,
                                    xpointlist=xpointlist_predict,
                                    ypointlist=ypointlist_predict,
                                    color_line=self.COLOR_MAP['traj_prediction_line'],
                                    line_wide=1.5,
                                    color_point=self.COLOR_MAP['traj_prediction_point'],
                                    size_point=3.0)
    
    
    def _plot_traj_line_points(self,xlinelist,ylinelist,xpointlist,ypointlist,color_line,line_wide,color_point,size_point):
        """绘制轨迹线 and 轨迹点.
        """
        self.axTraj.plot(
            xlinelist,
            ylinelist,
            color=color_line,
            lw=line_wide )
        self.axTraj.scatter(
            xpointlist,
            ypointlist,
            color=color_point,
            s=size_point )
         
            
    def _generate_float_list(self,start,step,count):
        """生成时间序列.
        """
        float_list = []
        float_list_tostr = []
        current_value = start
        for _ in range(count):
            float_list.append(round(current_value,2))
            float_list_tostr.append(str(round(current_value,2)))
            current_value += step
        return float_list,float_list_tostr
    
    
    def _generate_float_list_2(self,start,step,end):
        """生成时间序列,case 2:1s短时预测.
        """
        float_list = []
        float_list_tostr = []
        current_value = start
        while current_value <= end:
            float_list.append(round(current_value,2))
            float_list_tostr.append(str(round(current_value,2)))
            current_value += step
        return float_list,float_list_tostr
    
    
    def _find_xylist_from_vehi_traj(self,strkey_list,vehi_traj):
        """提取 5s的所有轨迹点 and 1 2 3 4 5秒处的轨迹点.
        """
        value_list_x = []
        value_list_y = []  
        for key in strkey_list:
            if key in vehi_traj:
                value_list_x.append(  vehi_traj[key]['x']  )
                value_list_y.append(  vehi_traj[key]['y']  )
        return value_list_x,value_list_y    


    def _plot_vehicles(self,observation:Observation) -> None:
        """绘制all车辆BOX.
        """
        for key,values in observation.vehicle_info.items():
            if key == 'ego':
                self._plot_single_vehicle(key,values,c=self.COLOR_MAP['ego_vehicle'])
            else:
                if values['shape']['vehicle_type'] == 'rock':
                    self._plot_single_rock(key, values, c=self.COLOR_MAP['static_obstacle'])  # 静态障碍物
                else:
                    self._plot_single_vehicle(key,values,c=self.COLOR_MAP['background_vehicle'])

    def _plot_single_rock(self, key: str, vehi: dict, c=None):
        """绘制单个车辆BOX.
        注:利用 matplotlib 和 patches 绘制小汽车,以 x 轴为行驶方向
        """
        x = vehi['x']
        y = vehi['y']
        yaw = vehi['yaw_rad']

        x_A3 = x - vehi['shape']['locationPoint2Rear'] * np.cos(yaw) + 0.5 * vehi['shape']['width'] * np.sin(yaw)
        y_A3 = y - vehi['shape']['locationPoint2Rear'] * np.sin(yaw) - 0.5 * vehi['shape']['width'] * np.cos(yaw)
        width_x = vehi['shape']['length']  # 绘图,x轴方向定义为宽
        height_y = vehi['shape']['width']  # 绘图,y轴方向定义为高

        self.axveh.add_patch(
            patches.Rectangle(
                xy=(x_A3, y_A3),  # 矩形的左下角坐标
                width=width_x,
                height=height_y,
                angle=yaw / np.pi * 180,
                color=c,
                fill=True,
                zorder=3))
        if key != 'ego':
            self.axveh.annotate('r', (x, y))

    def _plot_single_vehicle(self,key:str,vehi:dict,c=None):
        """绘制单个车辆BOX.
        注:利用 matplotlib 和 patches 绘制小汽车,以 x 轴为行驶方向
        """
        x = vehi['x']
        y = vehi['y']
        yaw = vehi['yaw_rad']
        
        x_A3 = x - vehi['shape']['locationPoint2Rear'] * np.cos(yaw) + 0.5* vehi['shape']['width'] *np.sin(yaw)
        y_A3 = y - vehi['shape']['locationPoint2Rear'] * np.sin(yaw) - 0.5* vehi['shape']['width'] *np.cos(yaw)
        width_x= vehi['shape']['length']  # 绘图,x轴方向定义为宽
        height_y= vehi['shape']['width']   # 绘图,y轴方向定义为高

        self.axveh.add_patch(
            patches.Rectangle(
                xy=(x_A3,y_A3),#矩形的左下角坐标
                width=width_x,
                height=height_y,
                angle=yaw / np.pi * 180,
                color=c,
                fill=True,
                zorder=3  ))
        if key != 'ego':
            self.axveh.annotate(key,(x,y))


    def _plot_warning_signal(self,observation:Observation ):
        """绘制主车碰撞时的提醒标志.
        """
        for key,values in observation.vehicle_info.items():
            if key == 'ego':
                x,y = [float(values[i]) for i in ['x','y']]
                self.axveh.scatter(x,y,s=60,c=self.COLOR_MAP['warning_signal'],alpha=1.0,marker=(8,1,30),zorder=4)


    def _plot_target_area(self,observation:Observation):
        """绘制该场景预设目标区域.
        """        
        if self.x_target and self.y_target:
            x,y = self.x_target,self.y_target
            codes_box = [MatplotlibPath.MOVETO] + [MatplotlibPath.LINETO] * 3 + [MatplotlibPath.CLOSEPOLY] 
            vertices_box = [(self.x_target[0],self.y_target[0]),
                            (self.x_target[1],self.y_target[1]),
                            (self.x_target[2],self.y_target[2]),
                            (self.x_target[3],self.y_target[3]),
                            (self.x_target[0],self.y_target[0])] # 4 points of polygon
            path_box = MatplotlibPath(vertices_box,codes_box)  # 定义对应Path
            pathpatch_box = patches.PathPatch(path_box,
                                              facecolor=self.COLOR_MAP['target_box_facecolor'],
                                              edgecolor='orangered',
                                              zorder=2,
                                              alpha=0.7)
            self.axbg.add_patch(pathpatch_box)
            if self.scenario_type == 'loading' and self.heading is not None:
                center_x,center_y = sum(x)/len(x),sum(y)/len(y)
                node_ux = 5*np.cos(self.heading[0])
                node_vy = 5*np.sin(self.heading[0])
                self.axbg.arrow(center_x, center_y, node_ux, node_vy, color='red', alpha=1, head_width=2.0, head_length=2.0)
    
    def _plot_traj_future_path(self, traj_future)->None:
        if traj_future == None:
            return
        
        for veh_id in traj_future.keys():
            x = []
            y = []
            for key, value in traj_future[veh_id]["traj_future_for_vis"].items():
                x.append(value['x'])
                y.append(value['y'])
            plt.plot(x, y, marker='.', linestyle='-', color='b', linewidth=1, markersize=1)
            
    def _plot_hybrid_astar_path(self, HA_path)->None:
        if HA_path == None:
            # print("无混合A*路径传入，不进行绘制")
            return
        x = [row[0] for row in HA_path]
        y = [row[1] for row in HA_path]
        plt.plot(x, y, marker='.', linestyle='-', color='r', linewidth=2, markersize=2)
        
            
    def _plot_hdmaps(self,observation:Observation,) -> None:
        """根据observation绘制地图.
        注:只要完成绘制工作即可.plt.plot().其他plt.show()之类的不需要添加.

        Args:
            observation (Observation): 当前时刻的观察值
        """
        if not observation.hdmaps:
            return
        # plot mask图,plot 道路片段、交叉口多边形划分;
        my_patch = (self.x_min,self.y_min ,self.x_max,self.y_max)  #  (x_min,y_min,x_max,y_max).
        # layer_names = ['intersection','road']
        layer_names =  ['road','intersection','loading_area','unloading_area','road_block']
        self._plot_hdmaps_render_map_patch(tgsc_map_explorer=observation.hdmaps['tgsc_map'].explorer,  # TgScenesMapExplorer类
                                           box_coords=my_patch,layer_names=layer_names,alpha=0.7,  # 坐标范围，需要绘制的层，渗透率
                                           render_egoposes_range=False,render_legend=False,
                                           bitmap=observation.hdmaps['image_mask'],  # Bitmap类
                                           # bitmap=observation.hdmaps['image_rgb'],
                                           ax= self.axbg)
        
    def _plot_battlefield_hdmaps_render_map_patch(self,
                                    box_coords:Tuple[float,float,float,float]=(0,0,1,1),
                                    layer_names:List[str] = None,
                                    alpha:float = 0.5,
                                    render_egoposes_range:bool = False,
                                    render_legend:bool = False,
                                    bitmap:Optional[BitMap] = None,
                                    ax:Axes = None) -> None:    
        x_min,y_min,x_max,y_max = box_coords
        local_width = x_max - x_min
        local_height = y_max - y_min

        # 渲染位图
        if bitmap is not None:
            if bitmap.bitmap_type == 'bitmap_mask':
                bitmap.render_mask_map_using_image_ndarray_local(ax,window_size=5,gray_flag=True) #!mask图降采样绘图
            elif bitmap.bitmap_type == 'bitmap_rgb':
                bitmap.render_rgb_map(ax)
            else:
                raise Exception('###Exception### 非法的 bitmap type:%s' % self.bitmap_type) # 自定义异常
        if render_egoposes_range:
            ax.add_patch(Rectangle((x_min,y_min),local_width,local_height,fill=False,linestyle='-.',color='red',lw=3))
            ax.text(x_min + local_width / 100,y_min + local_height / 2,"%g m" % local_height,
                    color='red',fontsize=14,weight='bold')
            ax.text(x_min + local_width / 2,y_min + local_height / 100,"%g m" % local_width,
                    color='red',fontsize=14,weight='bold') #fig上添加文字标注
        if render_legend:
            ax.legend(frameon=True,loc='upper right')
  
    def _plot_hdmaps_render_map_patch(self,
                                    tgsc_map_explorer:TgScenesMapExplorer =None,
                                    box_coords:Tuple[float,float,float,float]=(0,0,1,1),
                                    layer_names:List[str] = None,
                                    alpha:float = 0.5,
                                    render_egoposes_range:bool = False,
                                    render_legend:bool = False,
                                    bitmap:Optional[BitMap] = None,
                                    ax:Axes = None) -> None:                        
        """ 渲染一个矩形图框,指定矩形框的xy坐标范围.By default renders all layers.        
        """
        x_min,y_min,x_max,y_max = box_coords
        local_width = x_max - x_min
        local_height = y_max - y_min

        # 渲染位图
        if bitmap is not None:
            if bitmap.bitmap_type == 'bitmap_mask':
                bitmap.render_mask_map_using_image_ndarray_local(ax,window_size=5,gray_flag=False) #!mask图降采样绘图
            elif bitmap.bitmap_type == 'bitmap_rgb':
                bitmap.render_rgb_map(ax)
            else:
                raise Exception('###Exception### 非法的 bitmap type:%s' % self.bitmap_type) # 自定义异常
            
        for layer_name in layer_names:  # 渲染各图层（5个）
            tgsc_map_explorer._render_layer(ax,layer_name,alpha)
         
        #  渲染intersection的reference path,lane（交叉口中）
        if self.flag_visilize_ref_path == True:
            tgsc_map_explorer.render_connector_path_centerlines(ax,alpha,resampling_rate=0.05)  # 20个点去一个绘图
        
        #  渲染road的reference path,lane（道路内）
        if self.flag_visilize_ref_path == True:
            tgsc_map_explorer.render_base_path_centerlines(ax,alpha,resampling_rate=0.2)  # 5个点取一个绘图
        
        if render_egoposes_range:
            ax.add_patch(Rectangle((x_min,y_min),local_width,local_height,fill=False,linestyle='-.',color='red',lw=3))
            ax.text(x_min + local_width / 100,y_min + local_height / 2,"%g m" % local_height,
                    color='red',fontsize=14,weight='bold')
            ax.text(x_min + local_width / 2,y_min + local_height / 100,"%g m" % local_width,
                    color='red',fontsize=14,weight='bold') #fig上添加文字标注
        if render_legend:
            ax.legend(frameon=True,loc='upper right')
  
  


import os
import setuptools
from setuptools import setup, find_packages

# 允许setup.py在任何路径下执行
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setuptools.setup(
    name="onsite_unstructured-lp",  # 库名
    version="0.1.0",  # 版本号
    author="kaiwen",  # 作者
    author_email="xhh666@sjtu.edu.cn",  # 作者邮箱
    description="A small example package",  # 简介
    long_description="long_description",  # 详细描述
    long_description_content_type="text/markdown",  # 描述语法
    url="https://github.com/pypa/sampleproject",  # 项目主页
    packages=find_packages(),
    package_data={
        "onsite_unstructured": [
            "onsite-unstructured/data/*.npy",
            "onsite-unstructured/kinetic_model/linux/*.slxp",  # 递归包含所有 .slxp 文件
            "onsite-unstructured/kinetic_model/linux/*.mat",   # 递归包含所有 .mat 文件
            "onsite-unstructured/kinetic_model/linux/*.m",     # 递归包含所有 .m 文件
            "onsite-unstructured/kinetic_model/linux/*.slx"    # 递归包含所有 .slx 文件
            "onsite-unstructured/kinetic_model/linux/*.slxc"    # 递归包含所有 .slx 文件

            "onsite-unstructured/kinetic_model/win/*.slxp",  # 递归包含所有 .slxp 文件
            "onsite-unstructured/kinetic_model/win/*.mat",   # 递归包含所有 .mat 文件
            "onsite-unstructured/kinetic_model/win/*.m",     # 递归包含所有 .m 文件
            "onsite-unstructured/kinetic_model/win/*.slx"    # 递归包含所有 .slx 文件
            "onsite-unstructured/kinetic_model/win/*.slxc"    # 递归包含所有 .slxc 文件
            "onsite-unstructured/kinetic_model/win/*.mexw64"    # 递归包含所有 .slx 文件

        ]
    },
    include_package_data=True,  # 激活 MANIFEST.in 文件
    classifiers=[  # 指定库的分类器
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[  # 依赖库
        'pyautogui',
        'Django >= 1.11',
    ],
    python_requires='>=3.6',
)